from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from contextlib import contextmanager

from toga.handlers import wrapped_handler

from .base import Widget


class Slider(Widget):
    def __init__(
        self,
        id: str | None = None,
        style=None,
        value: float | None = None,
        min: float = None,  # Default to 0.0 when range is removed
        max: float = None,  # Default to 1.0 when range is removed
        tick_count: int | None = None,
        on_change: callable | None = None,
        on_press: callable | None = None,
        on_release: callable | None = None,
        enabled: bool = True,
        range: tuple[float, float] | None = None,  # DEPRECATED
    ):
        """Create a new Slider widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param value: Initial :any:`value` of the slider. Defaults to the mid-point of
            the range.
        :param min: Initial minimum value of the slider. Defaults to 0.
        :param max: Initial maximum value of the slider. Defaults to 1.
        :param tick_count: Initial :any:`tick_count` for the slider. If :any:`None`, the
            slider will be continuous.
        :param on_change: Initial :any:`on_change` handler.
        :param on_press: Initial :any:`on_press` handler.
        :param on_release: Initial :any:`on_release` handler.
        :param enabled: Whether the user can interact with the widget.
        :param range: **DEPRECATED**; use ``min`` and ``max`` instead. Initial
            :any:`range` of the slider. Defaults to ``(0, 1)``.
        """
        super().__init__(id=id, style=style)
        self._impl = self.factory.Slider(interface=self)

        ######################################################################
        # 2023-06: Backwards compatibility
        ######################################################################
        if range is not None:
            if min is not None or max is not None:
                raise ValueError(
                    "range cannot be specified if min and max are specified"
                )
            else:
                warnings.warn(
                    "Slider.range has been deprecated in favor of Slider.min and Slider.max",
                    DeprecationWarning,
                )
                min, max = range
        else:
            # This provides defaults values for min/max.
            if min is None:
                min = 0.0
            if max is None:
                max = 1.0
        ######################################################################
        # End backwards compatibility
        ######################################################################

        # Set a dummy handler before installing the actual on_change, because we do not want
        # on_change triggered by the initial value being set
        self.on_change = None
        self.min = min
        self.max = max
        self.tick_count = tick_count
        if value is None:
            value = (min + max) / 2
        self.value = value

        self.on_change = on_change
        self.on_press = on_press
        self.on_release = on_release

        self.enabled = enabled

    _MIN_WIDTH = 100

    # Backends are inconsistent about when they produce events for programmatic changes,
    # so we deal with those in the interface layer.
    @contextmanager
    def _programmatic_change(self):
        old_value = self.value
        on_change = self._on_change
        self.on_change = None
        yield old_value

        self._on_change = on_change
        if self.value != old_value:
            on_change()

    @property
    def value(self) -> float:
        """Current value.

        If the slider is discrete, setting the value will round it to the nearest tick.

        :raises ValueError: If set to a value which is outside of the :any:`range`.
        """
        return self._impl.get_value()

    @value.setter
    def value(self, value):
        if value < self.min:
            value = self.min
        elif value > self.max:
            value = self.max
        with self._programmatic_change():
            self._set_value(value)

    def _set_value(self, value):
        self._impl.set_value(self._round_value(float(value)))

    def _round_value(self, value):
        step = self.tick_step
        if step is not None:
            # Round to the nearest tick.
            value = self.min + round((value - self.min) / step) * step
        return value

    @property
    def min(self) -> float:
        """Minimum allowed value.

        When setting this property, the current :attr:`value` and :attr:`max` will be
        clipped against the new minimum value.
        """
        return self._impl.get_min()

    @min.setter
    def min(self, value):
        with self._programmatic_change() as old_value:
            # Some backends will clip the current value within the range automatically,
            # but do it ourselves to be certain. In discrete mode, setting self.value also
            # rounds to the new positions of the ticks.
            _min = float(value)
            _max = self.max
            if _max < _min:
                _max = _min
                self._impl.set_max(_max)

            self._impl.set_min(_min)
            self._set_value(max(_min, min(_max, old_value)))

    @property
    def max(self) -> float:
        """Maximum allowed value.

        When setting this property, the current :attr:`value` and :attr:`min` will be
        clipped against the new maximum value.
        """
        return self._impl.get_max()

    @max.setter
    def max(self, value):
        with self._programmatic_change() as old_value:
            # Some backends will clip the current value within the range automatically,
            # but do it ourselves to be certain. In discrete mode, setting self.value also
            # rounds to the new positions of the ticks.
            _min = self.min
            _max = float(value)
            if _min > _max:
                _min = _max
                self._impl.set_min(_min)

            self._impl.set_max(_max)
            self._set_value(max(_min, min(_max, old_value)))

    @property
    def tick_count(self) -> int | None:
        """Number of tick marks to display on the slider.

        * If this is ``None``, the slider will be continuous.
        * If this is an ``int``, the slider will be discrete, and will have the given
          number of possible values, equally spaced within the :any:`range`.

        Setting this property to an ``int`` will round the current value to the nearest
        tick.

        :raises ValueError: If set to a count which is not at least 2 (for the min and
            max).

        .. note::

            On iOS, tick marks are not currently displayed, but discrete mode will
            otherwise work correctly.
        """
        return self._impl.get_tick_count()

    @tick_count.setter
    def tick_count(self, tick_count):
        if (tick_count is not None) and (tick_count < 2):
            raise ValueError("tick count must be at least 2")
        with self._programmatic_change() as old_value:
            # Some backends will round the current value to the nearest tick
            # automatically, but do it ourselves to be certain. Some backends also require
            # the value to be refreshed when moving between discrete and continuous mode,
            # because this causes a change in the native range.
            self._impl.set_tick_count(tick_count)
            self.value = old_value

    @property
    def tick_step(self) -> float | None:
        """Step between adjacent ticks.

        * If the slider is continuous, this property returns ``None``
        * If the slider is discrete, it returns the difference in value between adjacent
          ticks.

        This property is read-only, and depends on the values of :any:`tick_count` and
        :any:`range`.
        """
        if self.tick_count is None or self.max == self.min:
            return None
        return (self.max - self.min) / (self.tick_count - 1)

    @property
    def tick_value(self) -> int | None:
        """Value of the slider, measured in ticks.

        * If the slider is continuous, this property returns ``None``.
        * If the slider is discrete, it returns an integer between 1 (representing
          :any:`min`) and :any:`tick_count` (representing :any:`max`).

        :raises ValueError: If set to anything inconsistent with the rules above.
        """
        if self.tick_count is not None:
            return round((self.value - self.min) / self.tick_step) + 1
        else:
            return None

    @tick_value.setter
    def tick_value(self, tick_value):
        if self.tick_count is None:
            if tick_value is not None:
                raise ValueError("cannot set tick value when tick count is None")
        else:
            if tick_value is None:
                raise ValueError(
                    "cannot set tick value to None when tick count is not None"
                )
            self.value = self.min + (tick_value - 1) * self.tick_step

    @property
    def on_change(self) -> callable:
        """Handler to invoke when the value of the slider is changed, either by the user
        or programmatically.

        Setting the widget to its existing value will not call the handler.
        """
        return self._on_change

    @on_change.setter
    def on_change(self, handler):
        self._on_change = wrapped_handler(self, handler)

    @property
    def on_press(self) -> callable:
        """Handler to invoke when the user presses the slider before changing it."""
        return self._on_press

    @on_press.setter
    def on_press(self, handler):
        self._on_press = wrapped_handler(self, handler)

    @property
    def on_release(self) -> callable:
        """Handler to invoke when the user releases the slider after changing it."""
        return self._on_release

    @on_release.setter
    def on_release(self, handler):
        self._on_release = wrapped_handler(self, handler)

    ######################################################################
    # 2023-06: Backwards compatibility
    ######################################################################
    @property
    def range(self) -> tuple[float, float]:
        """**DEPRECATED**; use :any:`min` and :any:`max` instead.

        Range of allowed values, in the form (min, max).

        If the provided min is greater than the max, both values will assume the value
        of the max.

        If the current value is less than the provided ``min``, the current value will
        be clipped to the minimum value. If the current value is greater than the
        provided ``max``, the current value will be clipped to the maximum value.
        """
        warnings.warn(
            "Slider.range has been deprecated in favor of Slider.min and Slider.max",
            DeprecationWarning,
        )
        return (self.min, self.max)

    @range.setter
    def range(self, range):
        warnings.warn(
            "Slider.range has been deprecated in favor of Slider.min and Slider.max",
            DeprecationWarning,
        )
        _min, _max = range
        self.min = _min
        self.max = _max


class SliderImpl(ABC):
    @abstractmethod
    def get_value(self): ...

    @abstractmethod
    def set_value(self, value): ...

    @abstractmethod
    def get_min(self): ...

    @abstractmethod
    def set_min(self, value): ...

    @abstractmethod
    def get_max(self): ...

    @abstractmethod
    def set_max(self, value): ...

    @abstractmethod
    def get_tick_count(self): ...

    @abstractmethod
    def set_tick_count(self, tick_count): ...


class IntSliderImpl(SliderImpl):
    """Base class for implementations which use integer values."""

    # Number of steps to use to approximate a continuous slider.
    CONTINUOUS_MAX = 10000

    def __init__(self):
        super().__init__()

        # Dummy values used during initialization.
        self.value = 0
        self.min = 0
        self.max = 1
        self.discrete = False

    def get_value(self):
        return self.value

    def set_value(self, value):
        span = self.max - self.min
        self.set_int_value(
            0 if span == 0 else round((value - self.min) / span * self.get_int_max())
        )
        self.value = value  # Cache the original value so we can round-trip it.

    def get_min(self):
        return self.min

    def set_min(self, value):
        self.min = value

    def get_max(self):
        return self.max

    def set_max(self, value):
        self.max = value

    def get_tick_count(self):
        return (self.get_int_max() + 1) if self.discrete else None

    def set_tick_count(self, tick_count):
        if tick_count is None:
            self.discrete = False
            self.set_int_max(self.CONTINUOUS_MAX)
        else:
            self.discrete = True
            self.set_int_max(tick_count - 1)
        self.set_ticks_visible(self.discrete)

    # Instead of calling the event handler directly, implementations should call this
    # method.
    def on_change(self):
        span = self.max - self.min
        self.value = self.min + (self.get_int_value() / self.get_int_max() * span)
        self.interface.on_change()

    @abstractmethod
    def get_int_value(self): ...

    @abstractmethod
    def set_int_value(self, value): ...

    @abstractmethod
    def get_int_max(self): ...

    @abstractmethod
    def set_int_max(self, max): ...

    @abstractmethod
    def set_ticks_visible(self, visible): ...



from toga.style.pack import CENTER, COLUMN, LEFT, RIGHT, ROW, Pack

from ..utils import ExampleNode, ExampleViewport, assert_layout


def test_left():
    root = ExampleNode(
        "root",
        style=Pack(direction=COLUMN, alignment=LEFT, height=300),
        children=[
            ExampleNode("space_filler", style=Pack(width=540, height=100)),
            ExampleNode(
                "container",
                style=Pack(direction=ROW, padding_left=110, padding_right=120),
                children=[
                    ExampleNode(
                        "widget_a",
                        style=Pack(width=30, padding_left=32, padding_right=34),
                    ),
                    ExampleNode(
                        "widget_b",
                        style=Pack(width=36, padding_left=38, padding_right=40),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (540, 300),
        (640, 300),
        {
            "origin": (0, 0),
            "content": (640, 300),
            "children": [
                {"origin": (0, 0), "content": (540, 100)},
                {
                    "origin": (110, 100),
                    "content": (210, 200),
                    "children": [
                        {"origin": (142, 100), "content": (30, 200)},
                        {"origin": (244, 100), "content": (36, 200)},
                    ],
                },
            ],
        },
    )


def test_center():
    root = ExampleNode(
        "root",
        style=Pack(direction=COLUMN, alignment=CENTER, height=300),
        children=[
            ExampleNode("space_filler", style=Pack(width=540, height=100)),
            ExampleNode(
                "container",
                style=Pack(direction=ROW, padding_left=110, padding_right=120),
                children=[
                    ExampleNode(
                        "widget_a",
                        style=Pack(width=30, padding_left=32, padding_right=34),
                    ),
                    ExampleNode(
                        "widget_b",
                        style=Pack(width=36, padding_left=38, padding_right=40),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (540, 300),
        (640, 300),
        {
            "origin": (0, 0),
            "content": (640, 300),
            "children": [
                {"origin": (50, 0), "content": (540, 100)},
                {
                    "origin": (210, 100),
                    "content": (210, 200),
                    "children": [
                        {"origin": (242, 100), "content": (30, 200)},
                        {"origin": (344, 100), "content": (36, 200)},
                    ],
                },
            ],
        },
    )


def test_right():
    root = ExampleNode(
        "root",
        style=Pack(direction=COLUMN, alignment=RIGHT, height=300),
        children=[
            ExampleNode("space_filler", style=Pack(width=540, height=100)),
            ExampleNode(
                "container",
                style=Pack(direction=ROW, padding_left=110, padding_right=120),
                children=[
                    ExampleNode(
                        "widget_a",
                        style=Pack(width=30, padding_left=32, padding_right=34),
                    ),
                    ExampleNode(
                        "widget_b",
                        style=Pack(width=36, padding_left=38, padding_right=40),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (540, 300),
        (640, 300),
        {
            "origin": (0, 0),
            "content": (640, 300),
            "children": [
                {"origin": (100, 0), "content": (540, 100)},
                {
                    "origin": (310, 100),
                    "content": (210, 200),
                    "children": [
                        {"origin": (342, 100), "content": (30, 200)},
                        {"origin": (444, 100), "content": (36, 200)},
                    ],
                },
            ],
        },
    )


def test_no_padding():
    root = ExampleNode(
        "root",
        style=Pack(direction=COLUMN, alignment=CENTER, height=300),
        children=[
            ExampleNode("space_filler", style=Pack(width=540, height=100)),
            ExampleNode("widget", style=Pack(width=440)),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (540, 300),
        (640, 300),
        {
            "origin": (0, 0),
            "content": (640, 300),
            "children": [
                {"origin": (50, 0), "content": (540, 100)},
                {"origin": (100, 100), "content": (440, 200)},
            ],
        },
    )


def test_column_box():
    root = ExampleNode(
        "root",
        style=Pack(direction=COLUMN, alignment=CENTER),
        children=[
            ExampleNode("space_filler", style=Pack(width=430, height=100)),
            ExampleNode(
                "container",
                style=Pack(direction=COLUMN, padding_left=110, padding_right=120),
                children=[
                    ExampleNode(
                        "widget_a",
                        style=Pack(
                            width=100, height=30, padding_top=32, padding_bottom=34
                        ),
                    ),
                    ExampleNode(
                        "widget_b",
                        style=Pack(
                            width=100, height=36, padding_top=38, padding_bottom=40
                        ),
                    ),
                ],
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (430, 310),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {"origin": (105, 0), "content": (430, 100)},
                {
                    "origin": (110, 100),
                    "content": (410, 210),
                    "children": [
                        {"origin": (110, 132), "content": (100, 30)},
                        {"origin": (110, 234), "content": (100, 36)},
                    ],
                },
            ],
        },
    )



import pytest

import toga
from toga_dummy.utils import (
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


@pytest.fixture
def window():
    return toga.Window()


@pytest.fixture
def content1():
    return toga.Box()


@pytest.fixture
def content2():
    return toga.Box()


@pytest.fixture
def content3():
    return toga.Box()


@pytest.fixture
def splitcontainer(content1, content2):
    return toga.SplitContainer(content=[content1, content2])


def test_widget_created():
    """A scroll container can be created with no arguments."""
    splitcontainer = toga.SplitContainer()
    assert splitcontainer._impl.interface == splitcontainer
    assert_action_performed(splitcontainer, "create SplitContainer")

    assert splitcontainer.content == (None, None)
    assert splitcontainer.direction == toga.SplitContainer.VERTICAL


def test_widget_created_with_values(content1, content2):
    """A split container can be created with arguments."""
    splitcontainer = toga.SplitContainer(
        content=[content1, content2],
        direction=toga.SplitContainer.HORIZONTAL,
    )
    assert splitcontainer._impl.interface == splitcontainer
    assert_action_performed(splitcontainer, "create SplitContainer")

    assert splitcontainer.content == (content1, content2)
    assert splitcontainer.direction == toga.SplitContainer.HORIZONTAL

    # The content has been assigned to the widget
    assert_action_performed_with(
        splitcontainer,
        "set content",
        content=(content1._impl, content2._impl),
        flex=[1, 1],
    )

    # The split container has been refreshed
    assert_action_performed(splitcontainer, "refresh")


@pytest.mark.parametrize(
    "include_left, include_right",
    [
        (False, False),
        (False, True),
        (True, False),
        (True, True),
    ],
)
def test_assign_to_app(app, content1, content2, include_left, include_right):
    """If the widget is assigned to an app, the content is also assigned."""
    splitcontainer = toga.SplitContainer(
        content=[
            content1 if include_left else None,
            content2 if include_right else None,
        ]
    )

    # Split container is initially unassigned
    assert splitcontainer.app is None

    # Assign the split container to the app
    splitcontainer.app = app

    # Split container is on the app
    assert splitcontainer.app == app

    # Content is also on the app
    if include_left:
        assert content1.app == app

    if include_right:
        assert content2.app == app


def test_assign_to_app_no_content(app):
    """If the widget is assigned to an app, and there is no content, there's no
    error."""
    splitcontainer = toga.SplitContainer()

    # Scroll container is initially unassigned
    assert splitcontainer.app is None

    # Assign the scroll container to the app
    splitcontainer.app = app

    # Scroll container is on the app
    assert splitcontainer.app == app


@pytest.mark.parametrize(
    "include_left, include_right",
    [
        (False, False),
        (False, True),
        (True, False),
        (True, True),
    ],
)
def test_assign_to_window(window, content1, content2, include_left, include_right):
    """If the widget is assigned to a window, the content is also assigned."""
    splitcontainer = toga.SplitContainer(
        content=[
            content1 if include_left else None,
            content2 if include_right else None,
        ]
    )

    # Split container is initially unassigned
    assert splitcontainer.window is None

    # Assign the split container to the window
    splitcontainer.window = window

    # Split container is on the window
    assert splitcontainer.window == window

    # Content is also on the window
    if include_left:
        assert content1.window == window

    if include_right:
        assert content2.window == window


def test_assign_to_window_no_content(window):
    """If the widget is assigned to an app, and there is no content, there's no
    error."""
    splitcontainer = toga.SplitContainer()

    # Scroll container is initially unassigned
    assert splitcontainer.window is None

    # Assign the scroll container to the window
    splitcontainer.window = window

    # Scroll container is on the window
    assert splitcontainer.window == window


def test_disable_no_op(splitcontainer):
    """SplitContainer doesn't have a disabled state."""
    # Enabled by default
    assert splitcontainer.enabled

    # Try to disable the widget
    splitcontainer.enabled = False

    # Still enabled.
    assert splitcontainer.enabled


def test_focus_noop(splitcontainer):
    """Focus is a no-op."""

    splitcontainer.focus()
    assert_action_not_performed(splitcontainer, "focus")


@pytest.mark.parametrize(
    "include_left, include_right",
    [
        (False, False),
        (False, True),
        (True, False),
        (True, True),
    ],
)
def test_set_content_widgets(
    splitcontainer,
    content1,
    content2,
    content3,
    include_left,
    include_right,
):
    """Widget content can be set to a list of widgets."""
    splitcontainer.content = [
        content2 if include_left else None,
        content3 if include_right else None,
    ]

    assert_action_performed_with(
        splitcontainer,
        "set content",
        content=(
            content2._impl if include_left else None,
            content3._impl if include_right else None,
        ),
        flex=[1, 1],
    )

    # The split container has been refreshed
    assert_action_performed(splitcontainer, "refresh")


@pytest.mark.parametrize(
    "include_left, include_right",
    [
        (False, False),
        (False, True),
        (True, False),
        (True, True),
    ],
)
def test_set_content_flex(
    splitcontainer,
    content2,
    content3,
    include_left,
    include_right,
):
    """Widget content can be set to a list of widgets with flex values."""
    splitcontainer.content = [
        (content2 if include_left else None, 2),
        (content3 if include_right else None, 3),
    ]

    assert_action_performed_with(
        splitcontainer,
        "set content",
        content=(
            content2._impl if include_left else None,
            content3._impl if include_right else None,
        ),
        flex=[2, 3],
    )

    # The split container has been refreshed
    assert_action_performed(splitcontainer, "refresh")


@pytest.mark.parametrize(
    "include_left, include_right",
    [
        (False, False),
        (False, True),
        (True, False),
        (True, True),
    ],
)
def test_set_content_flex_mixed(
    splitcontainer,
    content2,
    content3,
    include_left,
    include_right,
):
    """Flex values will be defaulted if missing."""
    splitcontainer.content = [
        content2 if include_left else None,
        (content3 if include_right else None, 3),
    ]

    assert_action_performed_with(
        splitcontainer,
        "set content",
        content=(
            content2._impl if include_left else None,
            content3._impl if include_right else None,
        ),
        flex=[1, 3],
    )

    # The split container has been refreshed
    assert_action_performed(splitcontainer, "refresh")


@pytest.mark.parametrize(
    "content, message",
    [
        (
            None,
            r"SplitContainer content must be a sequence with exactly 2 elements",
        ),
        (
            [],
            r"SplitContainer content must be a sequence with exactly 2 elements",
        ),
        (
            [toga.Box()],
            r"SplitContainer content must be a sequence with exactly 2 elements",
        ),
        (
            [toga.Box(), toga.Box(), toga.Box()],
            r"SplitContainer content must be a sequence with exactly 2 elements",
        ),
        (
            [toga.Box(), (toga.Box(),)],
            r"An item in SplitContainer content must be a 2-tuple containing "
            r"the widget, and the flex weight to assign to that widget.",
        ),
        (
            [toga.Box(), (toga.Box(), 42, True)],
            r"An item in SplitContainer content must be a 2-tuple containing "
            r"the widget, and the flex weight to assign to that widget.",
        ),
        (
            [toga.Box(), (toga.Box(), 0)],
            r"The flex value for an item in a SplitContainer must be >0",
        ),
        (
            [toga.Box(), (toga.Box(), -1)],
            r"The flex value for an item in a SplitContainer must be >0",
        ),
    ],
)
def test_set_content_invalid(splitcontainer, content, message):
    """Widget content can only be set to valid values."""

    with pytest.raises(ValueError, match=message):
        splitcontainer.content = content


def test_direction(splitcontainer):
    """The direction of the splitcontainer can be changed."""

    splitcontainer.direction = toga.SplitContainer.HORIZONTAL

    # The direction has been set
    assert splitcontainer.direction == toga.SplitContainer.HORIZONTAL

    # The split container has been refreshed
    assert_action_performed(splitcontainer, "refresh")



from travertino.size import at_least

from toga.style.pack import COLUMN, Pack

from ..utils import ExampleNode, ExampleViewport, assert_layout


def test_tutorial_0():
    root = ExampleNode(
        "app",
        style=Pack(),
        children=[
            ExampleNode(
                "button", style=Pack(flex=1, padding=50), size=(at_least(120), 30)
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (220, 130),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [{"origin": (50, 50), "content": (540, 30)}],
        },
    )


def test_vertical():
    root = ExampleNode(
        "app",
        style=Pack(direction=COLUMN),
        children=[
            # ExampleNode('button', style=Pack(flex=1), size=(30, at_least(120))),
            ExampleNode(
                "button", style=Pack(flex=1, padding=50), size=(30, at_least(120))
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(480, 640))
    assert_layout(
        root,
        (130, 220),
        (480, 640),
        {
            "origin": (0, 0),
            "content": (480, 640),
            "children": [{"origin": (50, 50), "content": (30, 540)}],
        },
    )



from pathlib import Path
from unittest.mock import Mock

import pytest

import toga
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
    attribute_value,
)


@pytest.fixture
def button():
    return toga.Button("Test Button")


@pytest.fixture
def sample_icon(app):
    return toga.Icon("resources/sample")


TEST_TEXT_VALUES = [
    ("New Text", "New Text"),
    ("", ""),
    (None, ""),
    ("\u200B", ""),
    (12345, "12345"),
    ("Contains\nnewline", "Contains"),
]


@pytest.mark.parametrize("value, expected", TEST_TEXT_VALUES)
def test_button_created(value, expected):
    """A button can be created."""
    button = toga.Button(text=value)

    # Round trip the impl/interface
    assert button._impl.interface == button

    assert_action_performed(button, "create Button")
    assert button.text == expected
    assert button.icon is None


def test_icon_button_created(button, sample_icon):
    """A button can be created."""
    button = toga.Button(icon=sample_icon)

    # Round trip the impl/interface
    assert button._impl.interface == button
    assert_action_performed(button, "create Button")
    assert button.text == ""
    assert button.icon == sample_icon


@pytest.mark.parametrize(
    "text,icon",
    [
        ("Bad text", "/path/to/icon"),
        ("", "/path/to/icon"),
    ],
)
def test_button_both_content(text, icon):
    """A button cannot have both text and an icon."""
    with pytest.raises(ValueError, match=r"Cannot specify both text and an icon"):
        toga.Button(text=text, icon=icon)


@pytest.mark.parametrize("value, expected", TEST_TEXT_VALUES)
def test_button_text(button, sample_icon, value, expected):
    """The button label can be modified."""
    assert button.text == "Test Button"
    assert button.icon is None

    # Clear the event log
    EventLog.reset()

    button.text = value
    assert button.text == expected
    assert button.icon is None

    # test backend has the right values
    assert attribute_value(button, "text") == expected
    assert attribute_value(button, "icon") is None

    # A refresh was performed
    assert_action_performed(button, "refresh")
    EventLog.reset()

    # Change to an icon
    button.icon = sample_icon
    assert button.icon == sample_icon

    # test backend has the right values
    assert attribute_value(button, "text") == ""
    assert attribute_value(button, "icon") == sample_icon

    # A refresh was performed
    assert_action_performed(button, "refresh")


@pytest.mark.parametrize("construct", [True, False])
def test_button_icon(button, construct):
    """The button icon can be modified."""
    assert button.text == "Test Button"
    assert button.icon is None

    if construct:
        icon = toga.Icon("path/to/icon")
    else:
        icon = "path/to/icon"

    # Set the icon
    button.icon = icon
    assert isinstance(button.icon, toga.Icon)
    assert button.icon.path == Path("path/to/icon")

    # test backend has the right values
    assert attribute_value(button, "text") == ""
    assert attribute_value(button, "icon").path == Path("path/to/icon")

    # A refresh was performed
    assert_action_performed(button, "refresh")
    EventLog.reset()

    # Change back to text
    button.text = "New value"
    assert button.text == "New value"
    assert button.icon is None

    # test backend has the right values
    assert attribute_value(button, "text") == "New value"
    assert attribute_value(button, "icon") is None

    # A refresh was performed
    assert_action_performed(button, "refresh")


def test_button_icon_none(button):
    """The button icon can be modified."""
    assert button.text == "Test Button"
    assert button.icon is None

    icon = toga.Icon("path/to/icon")

    EventLog.reset()

    # Set the icon to None; the button is already a text button,
    # so this doesn't change the text label.
    button.icon = None

    assert button.text == "Test Button"
    assert button.icon is None

    # No refresh was performed
    assert_action_not_performed(button, "refresh")
    EventLog.reset()

    # Set the icon
    button.icon = icon
    assert isinstance(button.icon, toga.Icon)
    assert button.icon.path == Path("path/to/icon")

    # test backend has the right values
    assert attribute_value(button, "text") == ""
    assert attribute_value(button, "icon").path == Path("path/to/icon")

    # A refresh was performed
    assert_action_performed(button, "refresh")
    EventLog.reset()

    # Assign a None icon
    button.icon = None
    assert button.text == ""
    assert button.icon is None

    # test backend has the right values
    assert attribute_value(button, "text") == ""
    assert attribute_value(button, "icon") is None

    # A refresh was performed
    assert_action_performed(button, "refresh")
    EventLog.reset()

    # Set the icon to None again; the button is already an icon button,
    # so the label is still ""
    button.icon = None

    assert button.text == ""
    assert button.icon is None

    # No refresh was performed
    assert_action_not_performed(button, "refresh")


def test_button_on_press(button):
    """The on_press handler can be invoked."""
    # No handler initially
    assert button._on_press._raw is None

    # Define and set a new callback
    handler = Mock()

    button.on_press = handler

    assert button.on_press._raw == handler

    # Invoke the callback
    button._impl.simulate_press()

    # Callback was invoked
    handler.assert_called_once_with(button)



import warnings

from .app import App, DocumentApp, DocumentMainWindow, MainWindow

# Resources
from .colors import hsl, hsla, rgb, rgba
from .command import Command, Group
from .documents import Document
from .fonts import Font
from .icons import Icon
from .images import Image
from .keys import Key

# Types
from .types import LatLng

# Widgets
from .widgets.activityindicator import ActivityIndicator
from .widgets.base import Widget
from .widgets.box import Box
from .widgets.button import Button
from .widgets.canvas import Canvas
from .widgets.dateinput import DateInput, DatePicker
from .widgets.detailedlist import DetailedList
from .widgets.divider import Divider
from .widgets.imageview import ImageView
from .widgets.label import Label
from .widgets.mapview import MapPin, MapView
from .widgets.multilinetextinput import MultilineTextInput
from .widgets.numberinput import NumberInput
from .widgets.optioncontainer import OptionContainer, OptionItem
from .widgets.passwordinput import PasswordInput
from .widgets.progressbar import ProgressBar
from .widgets.scrollcontainer import ScrollContainer
from .widgets.selection import Selection
from .widgets.slider import Slider
from .widgets.splitcontainer import SplitContainer
from .widgets.switch import Switch
from .widgets.table import Table
from .widgets.textinput import TextInput
from .widgets.timeinput import TimeInput, TimePicker
from .widgets.tree import Tree
from .widgets.webview import WebView
from .window import Window


class NotImplementedWarning(RuntimeWarning):
    # pytest.warns() requires that Warning() subclasses are constructed by passing a
    # single argument (the warning message). Use a factory method to avoid reproducing
    # the message format and the warn invocation.
    @classmethod
    def warn(self, platform, feature):
        """Raise a warning that a feature isn't implemented on a platform."""
        warnings.warn(NotImplementedWarning(f"[{platform}] Not implemented: {feature}"))


__all__ = [
    "NotImplementedWarning",
    # Applications
    "App",
    "DocumentApp",
    "MainWindow",
    "DocumentMainWindow",
    # Commands
    "Command",
    "Group",
    # Documents
    "Document",
    # Keys
    "Key",
    # Resources
    "hsl",
    "hsla",
    "rgb",
    "rgba",
    "Font",
    "Icon",
    "Image",
    # Types
    "LatLng",
    # Widgets
    "ActivityIndicator",
    "Box",
    "Button",
    "Canvas",
    "DateInput",
    "DetailedList",
    "Divider",
    "ImageView",
    "Label",
    "MapPin",
    "MapView",
    "MultilineTextInput",
    "NumberInput",
    "OptionContainer",
    "OptionItem",
    "PasswordInput",
    "ProgressBar",
    "ScrollContainer",
    "Selection",
    "Slider",
    "SplitContainer",
    "Switch",
    "Table",
    "TextInput",
    "TimeInput",
    "Tree",
    "WebView",
    "Widget",
    "Window",
    # Deprecated widget names
    "DatePicker",
    "TimePicker",
]


def _package_version(file, name):
    try:
        # Read version from SCM metadata
        # This will only exist in a development environment
        from setuptools_scm import get_version

        # Excluded from coverage because a pure test environment (such as the one
        # used by tox in CI) won't have setuptools_scm
        return get_version(root="../../..", relative_to=file)  # pragma: no cover
    except (ModuleNotFoundError, LookupError):
        # If setuptools_scm isn't in the environment, the call to import will fail.
        # If it *is* in the environment, but the code isn't a git checkout (e.g.,
        # it's been pip installed non-editable) the call to get_version() will fail.
        # If either of these occurs, read version from the installer metadata.
        import importlib.metadata

        # The Toga package names as defined in setup.cfg all use dashes.
        package = "toga-core" if name == "toga" else name.replace("_", "-")
        return importlib.metadata.version(package)


__version__ = _package_version(__file__, __name__)



from __future__ import annotations

from typing import Any, Protocol

import toga
from toga.handlers import wrapped_handler

from .base import Widget


class MapPin:
    def __init__(
        self,
        location: toga.LatLng | tuple[float, float],
        *,
        title: str,
        subtitle: str | None = None,
    ):
        """Create a new map pin.

        :param location: A tuple describing the (latitude, longitude) for the pin.
        :param title: The title to apply to the pin.
        :param subtitle: A subtitle label to apply to the pin.
        """
        self._location = toga.LatLng(*location)
        self._title = title
        self._subtitle = subtitle

        # A pin isn't tied to a map at time of creation.
        self.interface = None
        self._native = None

    def __repr__(self):
        if self.subtitle:
            label = f"; {self.title} - {self.subtitle}"
        else:
            label = f"; {self.title}"

        return f"<MapPin @ {self.location}{label}>"

    @property
    def location(self) -> toga.LatLng:
        "The (latitude, longitude) where the pin is located."
        return self._location

    @location.setter
    def location(self, coord: toga.LatLng | tuple[float, float]) -> None:
        self._location = toga.LatLng(*coord)
        if self.interface:
            self.interface._impl.update_pin(self)

    @property
    def title(self) -> str:
        """The title of the pin."""
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        self._title = str(title)
        if self.interface:
            self.interface._impl.update_pin(self)

    @property
    def subtitle(self) -> str | None:
        "The subtitle of the pin."
        return self._subtitle

    @subtitle.setter
    def subtitle(self, subtitle: str | None) -> None:
        if subtitle is not None:
            subtitle = str(subtitle)
        self._subtitle = subtitle
        if self.interface:
            self.interface._impl.update_pin(self)


class MapPinSet:
    def __init__(self, interface, pins):
        self.interface = interface
        self._pins = set()

        if pins:
            for item in pins:
                self.add(item)

    def __repr__(self):
        return f"<MapPinSet ({len(self)} pins)>"

    def __iter__(self):
        """Return an iterator over the pins on the map."""
        return iter(self._pins)

    def __len__(self):
        """Return the number of pins being displayed."""
        return len(self._pins)

    def add(self, pin):
        """Add a new pin to the map.

        :param pin: The :any:`toga.MapPin` instance to add.
        """
        pin.interface = self.interface
        self._pins.add(pin)
        self.interface._impl.add_pin(pin)

    def remove(self, pin):
        """Remove a pin from the map.

        :param pin: The  :any:`toga.MapPin` instance to remove.
        """
        self.interface._impl.remove_pin(pin)
        self._pins.remove(pin)
        pin.interface = None

    def clear(self):
        """Remove all pins from the map."""
        for pin in self._pins:
            self.interface._impl.remove_pin(pin)
        self._pins = set()


class OnSelectHandler(Protocol):
    def __call__(self, widget: MapView, *, pin: MapPin, **kwargs: Any) -> None:
        """A handler that will be invoked when the user selects a map pin.

        :param widget: The button that was pressed.
        :param pin: The pin that was selected.
        :param kwargs: Ensures compatibility with arguments added in future versions.
        """
        ...


class MapView(Widget):
    def __init__(
        self,
        id=None,
        style=None,
        location: toga.LatLng | tuple[float, float] | None = None,
        zoom: int = 11,
        pins: list[MapPin] | None = None,
        on_select: toga.widgets.mapview.OnSelectHandler | None = None,
    ):
        """Create a new MapView widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        :param location: The initial latitude/longitude where the map should be
            centered. If not provided, the initial location for the map is Perth,
            Australia.
        :param zoom: The initial zoom level for the map.
        :param pins: The initial pins to display on the map.
        :param on_select: A handler that will be invoked when the user selects a map
            pin.
        """
        super().__init__(id=id, style=style)

        self._impl = self.factory.MapView(interface=self)

        self._pins = MapPinSet(self, pins)

        if location:
            self.location = location
        else:
            # Default location is Perth, Australia. Because why not?
            self.location = (-31.9559, 115.8606)

        self.zoom = zoom

        self.on_select = on_select

    @property
    def location(self) -> toga.LatLng:
        """The latitude/longitude where the map is centered.

        A tuple of ``(latitude, longitude)`` can be provided as input; this will be
        converted into a :any:`toga.LatLng` object.
        """
        return self._impl.get_location()

    @location.setter
    def location(self, coordinates: toga.LatLng | tuple[float, float]):
        self._impl.set_location(toga.LatLng(*coordinates))

    @property
    def zoom(self) -> int:
        """Set the zoom level for the map.

        The zoom level is an integer in the range 0-20 (inclusive). It can be used to
        set the number of degrees of longitude that will span a 256 :ref:`CSS pixel
        <css-units>` region in the horizontal axis of the map, following the
        relationship::

            longitude_per_256_pixels = 360 / (2**zoom)

        In practical terms, this means a 256px square will cover:

        * 0-2: Whole world
        * 3-6: Large countries
        * 7-8: Small countries, or a state in a large country
        * 9-11: The extent of a city
        * 12-14: Suburbs of a city, or small towns
        * 15-17: Roads at the level useful for navigation
        * 18-19: Individual buildings
        * 20: A single building

        These zoom levels use the same mathematical basis as the OpenStreetMap API. See
        `OpenStreetMap's documentation on zoom levels
        <https://wiki.openstreetmap.org/wiki/Zoom_levels>`__ for more details.

        If the provided zoom value is outside the supported range, it will be clipped.

        At very low zoom levels, some backends may constrain the viewable range to avoid
        repeating map tiles in the visible area. This effectively sets a minimum bound
        on the zoom level that can be requested. The value of this minimum varies
        depending on the size and aspect ratio of the map view.
        """
        return round(self._impl.get_zoom())

    @zoom.setter
    def zoom(self, value: int):
        value = int(value)
        if value < 0:
            value = 0
        elif value > 20:
            value = 20

        self._impl.set_zoom(value)

    @property
    def pins(self) -> MapPinSet:
        """The set of pins currently being displayed on the map"""
        return self._pins

    @property
    def on_select(self) -> toga.widgets.mapview.OnSelectHandler:
        """The handler to invoke when the user selects a pin on a map.

        **Note:** This is not currently supported on GTK or Windows.
        """
        return self._on_select

    @on_select.setter
    def on_select(self, handler: toga.widgets.mapview.OnSelectHandler | None):
        if handler and not getattr(self._impl, "SUPPORTS_ON_SELECT", True):
            self.factory.not_implemented("MapView.on_select")

        self._on_select = wrapped_handler(self, handler)



import random
from unittest.mock import Mock

import pytest

import toga
from toga.command import CommandSet, Separator


def test_create():
    """A CommandSet can be created with defaults."""
    cs = CommandSet()

    assert list(cs) == []
    assert cs.on_change is None


def test_create_with_values():
    """A CommandSet can be created with values."""
    change_handler = Mock()
    cs = CommandSet(on_change=change_handler)

    assert list(cs) == []
    assert cs.on_change == change_handler


@pytest.mark.parametrize("change_handler", [(None), (Mock())])
def test_add_clear(app, change_handler):
    """Commands can be added and removed from a commandset."""
    # Put some commands into the app
    cmd_a = toga.Command(None, text="App command a")
    cmd_b = toga.Command(None, text="App command b", order=10)
    app.commands.add(cmd_a, cmd_b)
    assert list(app.commands) == [cmd_a, cmd_b]

    # Create a standalone command set and add some commands
    cs = CommandSet(on_change=change_handler)
    cmd1a = toga.Command(None, text="Test command 1a", order=3)
    cmd1b = toga.Command(None, text="Test command 1b", order=1)
    cs.add(cmd1a, cmd1b)

    # Change handler was called once.
    if change_handler:
        change_handler.assert_called_once()
        change_handler.reset_mock()

    # Command set has commands, and the order is the opposite to the insertion order.
    assert list(cs) == [cmd1b, cmd1a]

    # New Commands aren't known to the app
    assert list(app.commands) == [cmd_a, cmd_b]

    # Clear the command set
    cs.clear()

    # Change handler was called once.
    if change_handler:
        change_handler.assert_called_once()
        change_handler.reset_mock()

    # Command set no commands.
    assert list(cs) == []

    # App command set hasn't changed.
    assert list(app.commands) == [cmd_a, cmd_b]


@pytest.mark.parametrize("change_handler", [(None), (Mock())])
def test_add_clear_with_app(app, change_handler):
    """Commands can be added and removed from a commandset that is linked to an app."""
    # Put some commands into the app
    cmd_a = toga.Command(None, text="App command a")
    cmd_b = toga.Command(None, text="App command b", order=10)
    app.commands.add(cmd_a, cmd_b)
    assert list(app.commands) == [cmd_a, cmd_b]

    # Create a command set that is linked to the app and add some commands
    cs = CommandSet(on_change=change_handler, app=app)
    cmd1a = toga.Command(None, text="Test command 1a", order=3)
    cmd1b = toga.Command(None, text="Test command 1b", order=1)
    cs.add(cmd1a, cmd1b)

    # Change handler was called once.
    if change_handler:
        change_handler.assert_called_once()
        change_handler.reset_mock()

    # Command set has commands, and the order is the opposite to the insertion order.
    assert list(cs) == [cmd1b, cmd1a]

    # New Commands are known to the app
    assert list(app.commands) == [cmd_a, cmd1b, cmd1a, cmd_b]

    # Add another command to the commandset
    cmd2 = toga.Command(None, text="Test command 2", order=2)
    cs.add(cmd2)

    # Change handler was called once.
    if change_handler:
        change_handler.assert_called_once()
        change_handler.reset_mock()

    # Command set has commands, and the output is ordered.
    assert list(cs) == [cmd1b, cmd2, cmd1a]

    # App also knows about the command
    assert list(app.commands) == [cmd_a, cmd1b, cmd2, cmd1a, cmd_b]

    # Clear the command set
    cs.clear()

    # Change handler was called once.
    if change_handler:
        change_handler.assert_called_once()
        change_handler.reset_mock()

    # Command set has no commands.
    assert list(cs) == []

    # App command set hasn't changed.
    assert list(app.commands) == [cmd_a, cmd1b, cmd2, cmd1a, cmd_b]


def test_ordering(
    parent_group_1,
    parent_group_2,
    child_group_1,
    child_group_2,
    child_group_3,
    child_group_4,
    child_group_5,
):
    """Ordering of groups, separators and commands is preserved."""

    # Menu structure is:
    # - parent group 1
    #   - Z                 - a group that starts with a command
    #   - child group 1
    #     - Y
    #     - X
    #     - W
    #     ---               - a separator between two items
    #     - B
    #   - V
    #   - child group 2
    #     - U
    #     - T
    #   - S
    # - parent group 2
    #   - child group 5     - a group that starts with a child group
    #     - G
    #   - A
    #   ---
    #   - child group 4     - a child group with a separator before it
    #     - E
    #   - child group 3     - a child group with a separator after it
    #     - D
    #   ---
    #   - F
    command_a = toga.Command(None, "A", group=parent_group_2, section=1, order=2)
    command_d = toga.Command(None, "D", group=child_group_3)
    command_e = toga.Command(None, "E", group=child_group_4)
    command_f = toga.Command(None, "F", group=parent_group_2, section=3)
    command_g = toga.Command(None, "G", group=child_group_5, section=1)
    command_b = toga.Command(None, "B", group=child_group_1, section=2, order=1)
    command_s = toga.Command(None, "S", group=parent_group_1, order=5)
    command_t = toga.Command(None, "T", group=child_group_2, order=2)
    command_u = toga.Command(None, "U", group=child_group_2, order=1)
    command_v = toga.Command(None, "V", group=parent_group_1, order=3)
    command_w = toga.Command(None, "W", group=child_group_1, order=4)
    command_x = toga.Command(None, "X", group=child_group_1, order=2)
    command_y = toga.Command(None, "Y", group=child_group_1, order=1)
    command_z = toga.Command(None, "Z", group=parent_group_1, order=1)

    commands = [
        command_z,
        command_y,
        command_x,
        command_w,
        command_b,
        command_v,
        command_u,
        command_t,
        command_s,
        command_e,
        command_f,
        command_d,
        command_a,
        command_g,
    ]

    # First iteration, use the order as defined. Then repeat multiple times
    # to validate insertion order doesn't matter.
    for attempt in range(0, 10):
        if attempt:
            random.shuffle(commands)
        cs = CommandSet()
        cs.add(*commands)

        assert list(cs) == [
            command_z,
            command_y,
            command_x,
            command_w,
            Separator(group=child_group_1),
            command_b,
            command_v,
            command_u,
            command_t,
            command_s,
            command_g,
            command_a,
            Separator(group=parent_group_2),
            command_e,
            command_d,
            Separator(group=parent_group_2),
            command_f,
        ]



from unittest.mock import Mock

import pytest

import toga
from toga_dummy.utils import assert_action_performed


def test_create(app):
    """A MainWindow can be created with minimal arguments."""
    window = toga.MainWindow()

    assert window.app == app
    assert window.content is None

    assert window._impl.interface == window
    assert_action_performed(window, "create Window")

    # We can't know what the ID is, but it must be a string.
    assert isinstance(window.id, str)
    # Window title is the app title.
    assert window.title == "Test App"
    assert window.position == (100, 100)
    assert window.size == (640, 480)
    assert window.resizable
    assert window.closable
    assert window.minimizable
    assert len(window.toolbar) == 0
    # No on-close handler
    assert window.on_close is None


def test_no_close():
    """An on_close handler cannot be set on MainWindow."""
    window = toga.MainWindow()

    with pytest.raises(
        ValueError,
        match=r"Cannot set on_close handler for the main window. Use the app on_exit handler instead.",
    ):
        window.on_close = Mock()



import pytest

from toga.colors import REBECCAPURPLE
from toga.style.pack import (
    BOLD,
    BOTTOM,
    CENTER,
    COLUMN,
    HIDDEN,
    ITALIC,
    JUSTIFY,
    LEFT,
    LTR,
    NONE,
    NORMAL,
    PACK,
    RIGHT,
    ROW,
    RTL,
    SMALL_CAPS,
    SYSTEM,
    SYSTEM_DEFAULT_FONT_SIZE,
    TOP,
    VISIBLE,
    Pack,
)


@pytest.mark.parametrize(
    "style, expected_css",
    [
        # Empty definition
        pytest.param(
            Pack(),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="empty",
        ),
        # Display
        pytest.param(
            Pack(display=PACK),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="display-pack",
        ),
        pytest.param(
            Pack(display=NONE),
            "display: none; flex-direction: row; flex: 0.0 0 auto;",
            id="display-none",
        ),
        # Visibility
        pytest.param(
            Pack(visibility=VISIBLE),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="visibility-visible",
        ),
        pytest.param(
            Pack(visibility=HIDDEN),
            "visibility: hidden; flex-direction: row; flex: 0.0 0 auto;",
            id="visibility-hidden",
        ),
        # Direction
        pytest.param(
            Pack(direction=ROW),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="direction-row",
        ),
        pytest.param(
            Pack(direction=COLUMN),
            "flex-direction: column; flex: 0.0 0 auto;",
            id="direction-column",
        ),
        # Width
        pytest.param(
            Pack(width=42),
            "flex-direction: row; width: 42px;",
            id="width-explicit",
        ),
        pytest.param(
            Pack(width=42, flex=5),
            "flex-direction: row; width: 42px;",
            id="width-explicit-flex",
        ),
        pytest.param(
            Pack(width=0),
            "flex-direction: row; width: 0px;",
            id="width-0",
        ),
        pytest.param(
            Pack(width=NONE),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="width-none",
        ),
        pytest.param(
            Pack(width=NONE, flex=5),
            "flex-direction: row; flex: 5.0 0 auto;",
            id="width-none-flex",
        ),
        pytest.param(
            Pack(direction=ROW, width=42),
            "flex-direction: row; width: 42px;",
            id="width-row-explicit",
        ),
        pytest.param(
            Pack(direction=ROW, width=42, flex=5),
            "flex-direction: row; width: 42px;",
            id="width-row-explicit-flex",
        ),
        pytest.param(
            Pack(direction=ROW, width=0),
            "flex-direction: row; width: 0px;",
            id="width-row-0",
        ),
        pytest.param(
            Pack(direction=ROW, width=NONE),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="width-row-none",
        ),
        pytest.param(
            Pack(direction=ROW, width=NONE, flex=5),
            "flex-direction: row; flex: 5.0 0 auto;",
            id="width-row-none-flex",
        ),
        pytest.param(
            Pack(direction=COLUMN, width=42),
            "flex-direction: column; flex: 0.0 0 auto; width: 42px;",
            id="width-column-explicit",
        ),
        pytest.param(
            Pack(direction=COLUMN, width=42, flex=5),
            "flex-direction: column; flex: 5.0 0 auto; width: 42px;",
            id="width-column-explicit-flex",
        ),
        pytest.param(
            Pack(direction=COLUMN, width=0),
            "flex-direction: column; flex: 0.0 0 auto; width: 0px;",
            id="width-column-0",
        ),
        pytest.param(
            Pack(direction=COLUMN, width=NONE),
            "flex-direction: column; flex: 0.0 0 auto;",
            id="width-column-none",
        ),
        pytest.param(
            Pack(direction=COLUMN, width=NONE, flex=5),
            "flex-direction: column; flex: 5.0 0 auto;",
            id="width-column-none-flex",
        ),
        # Height
        pytest.param(
            Pack(height=42),
            "flex-direction: row; flex: 0.0 0 auto; height: 42px;",
            id="height-explicit",
        ),
        pytest.param(
            Pack(height=42, flex=5),
            "flex-direction: row; flex: 5.0 0 auto; height: 42px;",
            id="height-explicit-flex",
        ),
        pytest.param(
            Pack(height=0),
            "flex-direction: row; flex: 0.0 0 auto; height: 0px;",
            id="height-0",
        ),
        pytest.param(
            Pack(height=NONE),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="height-none",
        ),
        pytest.param(
            Pack(height=NONE, flex=5),
            "flex-direction: row; flex: 5.0 0 auto;",
            id="height-none-flex",
        ),
        pytest.param(
            Pack(direction=ROW, height=42),
            "flex-direction: row; flex: 0.0 0 auto; height: 42px;",
            id="height-row-explicit",
        ),
        pytest.param(
            Pack(direction=ROW, height=42, flex=5),
            "flex-direction: row; flex: 5.0 0 auto; height: 42px;",
            id="height-row-explicit",
        ),
        pytest.param(
            Pack(direction=ROW, height=0),
            "flex-direction: row; flex: 0.0 0 auto; height: 0px;",
            id="height-row-0",
        ),
        pytest.param(
            Pack(direction=ROW, height=NONE),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="height-row-none",
        ),
        pytest.param(
            Pack(direction=ROW, height=NONE, flex=5),
            "flex-direction: row; flex: 5.0 0 auto;",
            id="height-row-none",
        ),
        pytest.param(
            Pack(direction=COLUMN, height=42),
            "flex-direction: column; height: 42px;",
            id="height-column-explicit",
        ),
        pytest.param(
            Pack(direction=COLUMN, height=42, flex=5),
            "flex-direction: column; height: 42px;",
            id="height-column-explicit-flex",
        ),
        pytest.param(
            Pack(direction=COLUMN, height=0),
            "flex-direction: column; height: 0px;",
            id="height-column-0",
        ),
        pytest.param(
            Pack(direction=COLUMN, height=NONE),
            "flex-direction: column; flex: 0.0 0 auto;",
            id="height-column-none",
        ),
        pytest.param(
            Pack(direction=COLUMN, height=NONE, flex=5),
            "flex-direction: column; flex: 5.0 0 auto;",
            id="height-column-none",
        ),
        # Alignment - default layout
        pytest.param(
            Pack(alignment=LEFT),
            "flex-direction: row; flex: 0.0 0 auto; align-items: start;",
            id="alignment-left",
        ),
        pytest.param(
            Pack(alignment=RIGHT),
            "flex-direction: row; flex: 0.0 0 auto; align-items: end;",
            id="alignment-right",
        ),
        pytest.param(  # alignment is ignored
            Pack(alignment=TOP),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="alignment-top",
        ),
        pytest.param(  # alignment is ignored
            Pack(alignment=BOTTOM),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="alignment-bottom",
        ),
        pytest.param(
            Pack(alignment=CENTER),
            "flex-direction: row; flex: 0.0 0 auto; align-items: center;",
            id="alignment-center",
        ),
        # Alignment - row layout
        pytest.param(
            Pack(direction=ROW, alignment=LEFT),
            "flex-direction: row; flex: 0.0 0 auto; align-items: start;",
            id="row-alignment-left",
        ),
        pytest.param(
            Pack(direction=ROW, alignment=RIGHT),
            "flex-direction: row; flex: 0.0 0 auto; align-items: end;",
            id="row-alignment-right",
        ),
        pytest.param(  # alignment is ignored
            Pack(direction=ROW, alignment=TOP),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="row-alignment-top",
        ),
        pytest.param(  # alignment is ignored
            Pack(direction=ROW, alignment=BOTTOM),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="row-alignment-bottom",
        ),
        pytest.param(
            Pack(direction=ROW, alignment=CENTER),
            "flex-direction: row; flex: 0.0 0 auto; align-items: center;",
            id="row-alignment-center",
        ),
        # Alignment - column layout
        pytest.param(  # alignment is ignored
            Pack(direction=COLUMN, alignment=LEFT),
            "flex-direction: column; flex: 0.0 0 auto;",
            id="column-alignment-left",
        ),
        pytest.param(  # alignment is ignored
            Pack(direction=COLUMN, alignment=RIGHT),
            "flex-direction: column; flex: 0.0 0 auto;",
            id="column-alignment-right",
        ),
        pytest.param(
            Pack(direction=COLUMN, alignment=TOP),
            "flex-direction: column; flex: 0.0 0 auto; align-items: start;",
            id="column-alignment-top",
        ),
        pytest.param(
            Pack(direction=COLUMN, alignment=BOTTOM),
            "flex-direction: column; flex: 0.0 0 auto; align-items: end;",
            id="column-alignment-bottom",
        ),
        pytest.param(
            Pack(direction=COLUMN, alignment=CENTER),
            "flex-direction: column; flex: 0.0 0 auto; align-items: center;",
            id="column-alignment-center",
        ),
        # Padding
        pytest.param(
            Pack(padding_top=42),
            "flex-direction: row; flex: 0.0 0 auto; margin-top: 42px;",
            id="padding-top",
        ),
        pytest.param(
            Pack(padding_bottom=42),
            "flex-direction: row; flex: 0.0 0 auto; margin-bottom: 42px;",
            id="padding-bottom",
        ),
        pytest.param(
            Pack(padding_left=42),
            "flex-direction: row; flex: 0.0 0 auto; margin-left: 42px;",
            id="padding-left",
        ),
        pytest.param(
            Pack(padding_right=42),
            "flex-direction: row; flex: 0.0 0 auto; margin-right: 42px;",
            id="padding-right",
        ),
        pytest.param(
            Pack(padding=42),
            (
                "flex-direction: row; flex: 0.0 0 auto; "
                "margin-top: 42px; margin-bottom: 42px; "
                "margin-left: 42px; margin-right: 42px;"
            ),
            id="padding",
        ),
        # Explicitly 0 padding
        pytest.param(
            Pack(padding_top=0),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="padding-top-0",
        ),
        pytest.param(
            Pack(padding_bottom=0),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="padding-bottom-0",
        ),
        pytest.param(
            Pack(padding_left=0),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="padding-left-0",
        ),
        pytest.param(
            Pack(padding_right=0),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="padding-right-0",
        ),
        pytest.param(
            Pack(padding=0),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="padding-0",
        ),
        # Color
        pytest.param(
            Pack(color=REBECCAPURPLE),
            "flex-direction: row; flex: 0.0 0 auto; color: rgb(102, 51, 153);",
            id="color",
        ),
        # Background Color
        pytest.param(
            Pack(background_color=REBECCAPURPLE),
            "flex-direction: row; flex: 0.0 0 auto; background-color: rgb(102, 51, 153);",
            id="background-color",
        ),
        # Text Alignment
        pytest.param(
            Pack(text_align=LEFT),
            "flex-direction: row; flex: 0.0 0 auto; text-align: left;",
            id="text-align-left",
        ),
        pytest.param(
            Pack(text_align=RIGHT),
            "flex-direction: row; flex: 0.0 0 auto; text-align: right;",
            id="text-align-right",
        ),
        pytest.param(
            Pack(text_align=CENTER),
            "flex-direction: row; flex: 0.0 0 auto; text-align: center;",
            id="text-align-center",
        ),
        pytest.param(
            Pack(text_align=JUSTIFY),
            "flex-direction: row; flex: 0.0 0 auto; text-align: justify;",
            id="text-align-justify",
        ),
        # Text Direction
        pytest.param(
            Pack(text_direction=RTL),
            "flex-direction: row; flex: 0.0 0 auto; text-direction: rtl;",
            id="text-align-rtl",
        ),
        pytest.param(
            Pack(text_direction=LTR),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="text-align-ltr",
        ),
        # Font
        pytest.param(
            Pack(font_family="Helvetica"),
            "flex-direction: row; flex: 0.0 0 auto; font-family: Helvetica;",
            id="font-family",
        ),
        pytest.param(
            Pack(font_family="Times New Roman"),
            'flex-direction: row; flex: 0.0 0 auto; font-family: "Times New Roman";',
            id="font-family",
        ),
        pytest.param(
            Pack(font_family=SYSTEM),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="font-family-explicit-default",
        ),
        pytest.param(
            Pack(font_style=ITALIC),
            "flex-direction: row; flex: 0.0 0 auto; font-style: italic;",
            id="font-style",
        ),
        pytest.param(
            Pack(font_style=NORMAL),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="font-style-explicit-default",
        ),
        pytest.param(
            Pack(font_weight=BOLD),
            "flex-direction: row; flex: 0.0 0 auto; font-weight: bold;",
            id="font-weight",
        ),
        pytest.param(
            Pack(font_weight=NORMAL),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="font-weight-explicit-default",
        ),
        pytest.param(
            Pack(font_variant=SMALL_CAPS),
            "flex-direction: row; flex: 0.0 0 auto; font-variant: small-caps;",
            id="font-variant",
        ),
        pytest.param(
            Pack(font_variant=NORMAL),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="font-variant-explicit-default",
        ),
        pytest.param(
            Pack(font_size=42),
            "flex-direction: row; flex: 0.0 0 auto; font-size: 42pt;",
            id="font-size",
        ),
        pytest.param(
            Pack(font_size=SYSTEM_DEFAULT_FONT_SIZE),
            "flex-direction: row; flex: 0.0 0 auto;",
            id="font-size-explicit-default",
        ),
        pytest.param(
            Pack(
                font_family="Helvetica",
                font_style=ITALIC,
                font_weight=BOLD,
                font_variant=SMALL_CAPS,
                font_size=42,
            ),
            (
                "flex-direction: row; flex: 0.0 0 auto; "
                "font-family: Helvetica; "
                "font-size: 42pt; "
                "font-weight: bold; "
                "font-style: italic; "
                "font-variant: small-caps;"
            ),
            id="font-full",
        ),
    ],
)
def test_rendering(style, expected_css):
    """An empty style node can be rendered."""
    assert style.__css__() == expected_css



from PIL.Image import Image as PILImage

from toga.images import Image as TogaImage
from toga_dummy.utils import assert_action_performed


def test_name(app):
    """The name of the screens can be retrieved."""
    assert app.screens[0].name == "Primary Screen"
    assert app.screens[1].name == "Secondary Screen"


def test_origin(app):
    """The origin of the screens can be retrieved."""
    assert app.screens[0].origin == (0, 0)
    assert app.screens[1].origin == (-1366, -768)


def test_size(app):
    """The size of the screens can be retrieved."""
    assert app.screens[0].size == (1920, 1080)
    assert app.screens[1].size == (1366, 768)


def test_as_image(app):
    """A screen can be captured as an image."""
    for screen in app.screens:
        # `as_image()` should default to `toga.images.Image` as format.
        toga_image_screenshot = screen.as_image()
        assert_action_performed(screen, "get image data")

        # Check if returned image is of type `toga.images.Image`, and the right size
        assert isinstance(toga_image_screenshot, TogaImage)
        assert toga_image_screenshot.size == screen.size


def test_as_image_format(app):
    """A screen can be captured as an image in a non-default format."""
    for screen in app.screens:
        # Capture image in PIL format
        pil_screenshot = screen.as_image(format=PILImage)
        assert_action_performed(screen, "get image data")

        # Check if returned image is of type `PIL.Image.Image`, and the right size
        assert isinstance(pil_screenshot, PILImage)
        assert pil_screenshot.size == screen.size



from unittest.mock import Mock

import pytest

import toga
from toga.sources import TreeSource
from toga_dummy.utils import (
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


@pytest.fixture
def on_select_handler():
    return Mock()


@pytest.fixture
def on_activate_handler():
    return Mock()


@pytest.fixture
def source():
    source = TreeSource(
        data={
            ("group1", 1, "A**"): [
                (
                    {"key": "A first", "value": 110, "other": "AA*"},
                    None,
                ),
                (
                    {"key": "A second", "value": 120, "other": "AB*"},
                    [],
                ),
                (
                    {"key": "A third", "value": 130, "other": "AC*"},
                    [
                        ({"key": "A third-first", "value": 131, "other": "ACA"}, None),
                        ({"key": "A third-second", "value": 132, "other": "ACB"}, None),
                    ],
                ),
            ],
            ("group2", 2, "B**"): [
                (
                    {"key": "B first", "value": 210, "other": "BA*"},
                    None,
                ),
                (
                    {"key": "B second", "value": 220, "other": "BB*"},
                    [],
                ),
                (
                    {"key": "B third", "value": 230, "other": "BC*"},
                    [
                        ({"key": "B third-first", "value": 231, "other": "BCA"}, None),
                        ({"key": "B third-second", "value": 232, "other": "BCB"}, None),
                    ],
                ),
            ],
        },
        accessors=["key", "value"],
    )
    return source


@pytest.fixture
def tree(source, on_select_handler, on_activate_handler):
    return toga.Tree(
        ["Title", "Value"],
        accessors=["key", "value"],
        data=source,
        on_select=on_select_handler,
        on_activate=on_activate_handler,
    )


def test_tree_created():
    """A minimal Tree can be created."""
    tree = toga.Tree(["First", "Second"])
    assert tree._impl.interface is tree
    assert_action_performed(tree, "create Tree")

    assert len(tree.data) == 0
    assert tree.headings == ["First", "Second"]
    assert tree.accessors == ["first", "second"]
    assert not tree.multiple_select
    assert tree.missing_value == ""
    assert tree.on_select._raw is None
    assert tree.on_activate._raw is None


def test_create_with_values(source, on_select_handler, on_activate_handler):
    """A Tree can be created with initial values."""
    tree = toga.Tree(
        ["First", "Second"],
        data=source,
        accessors=["primus", "secondus"],
        multiple_select=True,
        on_select=on_select_handler,
        on_activate=on_activate_handler,
        missing_value="Boo!",
    )
    assert tree._impl.interface == tree
    assert_action_performed(tree, "create Tree")

    assert len(tree.data) == 2
    assert tree.headings == ["First", "Second"]
    assert tree.accessors == ["primus", "secondus"]
    assert tree.multiple_select
    assert tree.missing_value == "Boo!"
    assert tree.on_select._raw == on_select_handler
    assert tree.on_activate._raw == on_activate_handler


def test_create_with_accessor_overrides():
    """A Tree can partially override accessors."""
    tree = toga.Tree(
        ["First", "Second"],
        accessors={"First": "override"},
    )
    assert tree._impl.interface == tree
    assert_action_performed(tree, "create Tree")

    assert len(tree.data) == 0
    assert tree.headings == ["First", "Second"]
    assert tree.accessors == ["override", "second"]


def test_create_no_headings():
    """A Tree can be created with no headings."""
    tree = toga.Tree(
        headings=None,
        accessors=["primus", "secondus"],
    )
    assert tree._impl.interface == tree
    assert_action_performed(tree, "create Tree")

    assert len(tree.data) == 0
    assert tree.headings is None
    assert tree.accessors == ["primus", "secondus"]


def test_create_headings_required():
    """A Tree requires either headings can be created with no headings."""
    with pytest.raises(
        ValueError,
        match=r"Cannot create a tree without either headings or accessors",
    ):
        toga.Tree()


def test_disable_no_op(tree):
    """Tree doesn't have a disabled state."""
    # Enabled by default
    assert tree.enabled

    # Try to disable the widget
    tree.enabled = False

    # Still enabled.
    assert tree.enabled


def test_focus_noop(tree):
    """Focus is a no-op."""

    tree.focus()
    assert_action_not_performed(tree, "focus")


@pytest.mark.parametrize(
    "data, all_attributes, extra_attributes",
    [
        # Dictionary of single values
        (
            {
                "People": {
                    "Alice": None,
                    "Bob": None,
                    "Charlie": None,
                }
            },
            False,
            False,
        ),
        # Dictionary of tuples
        (
            {
                ("People", None, None): {
                    ("Alice", 123, "extra1"): None,
                    ("Bob", 234, "extra2"): None,
                    ("Charlie", 345, "extra4"): None,
                }
            },
            True,
            False,
        ),
        # List of tuples with tuples
        (
            [
                (
                    ("People", None, None),
                    [
                        (("Alice", 123, "extra1"), None),
                        (("Bob", 234, "extra2"), None),
                        (("Charlie", 345, "extra3"), None),
                    ],
                ),
            ],
            True,
            False,
        ),
        # List of tuples with Dictionaries
        (
            [
                (
                    {"key": "People"},
                    [
                        ({"key": "Alice", "value": 123, "other": "extra1"}, None),
                        ({"key": "Bob", "value": 234, "other": "extra2"}, None),
                        ({"key": "Charlie", "value": 345, "other": "extra3"}, None),
                    ],
                ),
            ],
            True,
            True,
        ),
    ],
)
def test_set_data(tree, on_select_handler, data, all_attributes, extra_attributes):
    """Data can be set from a variety of sources."""

    # The selection hasn't changed yet.
    on_select_handler.assert_not_called()

    # Change the data
    tree.data = data

    # This triggered the select handler
    on_select_handler.assert_called_once_with(tree)

    # A TreeSource has been constructed
    assert isinstance(tree.data, TreeSource)
    assert len(tree.data) == 1
    assert len(tree.data[0]) == 3

    # The accessors are mapped in order.
    assert tree.data[0].key == "People"

    assert tree.data[0][0].key == "Alice"
    assert tree.data[0][1].key == "Bob"
    assert tree.data[0][2].key == "Charlie"

    if all_attributes:
        assert tree.data[0][0].value == 123
        assert tree.data[0][1].value == 234
        assert tree.data[0][2].value == 345

    if extra_attributes:
        assert tree.data[0][0].other == "extra1"
        assert tree.data[0][1].other == "extra2"
        assert tree.data[0][2].other == "extra3"


def test_single_selection(tree, on_select_handler):
    """The current selection can be retrieved."""
    # Selection is initially empty
    assert tree.selection is None
    on_select_handler.assert_not_called()

    # Select an item
    tree._impl.simulate_selection((0, 1))

    # Selection returns a single row
    assert tree.selection == tree.data[0][1]

    # Selection handler was triggered
    on_select_handler.assert_called_once_with(tree)


def test_multiple_selection(source, on_select_handler):
    """A multi-select tree can have the selection retrieved."""
    tree = toga.Tree(
        ["Title", "Value"],
        data=source,
        multiple_select=True,
        on_select=on_select_handler,
    )
    # Selection is initially empty
    assert tree.selection == []
    on_select_handler.assert_not_called()

    # Select an item
    tree._impl.simulate_selection([(0, 1), (1, 2, 1)])

    # Selection returns a list of rows
    assert tree.selection == [tree.data[0][1], tree.data[1][2][1]]

    # Selection handler was triggered
    on_select_handler.assert_called_once_with(tree)


def test_expand_collapse(tree):
    """The rows on a tree can be expanded and collapsed."""

    # Expand the full tree
    tree.expand()
    assert_action_performed_with(tree, "expand all")

    # Collapse a single node
    tree.collapse(tree.data[1][2])
    assert_action_performed_with(tree, "collapse node", node=tree.data[1][2])

    # Expand a single node
    tree.expand(tree.data[1][2])
    assert_action_performed_with(tree, "expand node", node=tree.data[1][2])

    # Collapse a leaf node
    tree.collapse(tree.data[1][2][1])
    assert_action_performed_with(tree, "collapse node", node=tree.data[1][2][1])

    # Expand a leaf node
    tree.expand(tree.data[1][2][1])
    assert_action_performed_with(tree, "expand node", node=tree.data[1][2][1])

    # Collapse the full tree
    tree.collapse()
    assert_action_performed_with(tree, "collapse all")


def test_activation(tree, on_activate_handler):
    """A row can be activated."""

    # Activate an item
    tree._impl.simulate_activate((0, 1))

    # Activate handler was triggered; the activated node is provided
    on_activate_handler.assert_called_once_with(tree, node=tree.data[0][1])


def test_insert_column_accessor(tree):
    """A column can be inserted at an accessor."""
    tree.insert_column("value", "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        tree,
        "insert column",
        index=1,
        heading="New Column",
        accessor="extra",
    )
    assert tree.headings == ["Title", "New Column", "Value"]
    assert tree.accessors == ["key", "extra", "value"]


def test_insert_column_unknown_accessor(tree):
    """If the insertion index accessor is unknown, an error is raised."""
    with pytest.raises(ValueError, match=r"'unknown' is not in list"):
        tree.insert_column("unknown", "New Column", accessor="extra")


def test_insert_column_index(tree):
    """A column can be inserted."""

    tree.insert_column(1, "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        tree,
        "insert column",
        index=1,
        heading="New Column",
        accessor="extra",
    )
    assert tree.headings == ["Title", "New Column", "Value"]
    assert tree.accessors == ["key", "extra", "value"]


def test_insert_column_big_index(tree):
    """A column can be inserted at an index bigger than the number of columns."""

    tree.insert_column(100, "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        tree,
        "insert column",
        index=2,
        heading="New Column",
        accessor="extra",
    )
    assert tree.headings == ["Title", "Value", "New Column"]
    assert tree.accessors == ["key", "value", "extra"]


def test_insert_column_negative_index(tree):
    """A column can be inserted at a negative index."""

    tree.insert_column(-2, "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        tree,
        "insert column",
        index=0,
        heading="New Column",
        accessor="extra",
    )
    assert tree.headings == ["New Column", "Title", "Value"]
    assert tree.accessors == ["extra", "key", "value"]


def test_insert_column_big_negative_index(tree):
    """A column can be inserted at a negative index larger than the number of
    columns."""

    tree.insert_column(-100, "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        tree,
        "insert column",
        index=0,
        heading="New Column",
        accessor="extra",
    )
    assert tree.headings == ["New Column", "Title", "Value"]
    assert tree.accessors == ["extra", "key", "value"]


def test_insert_column_no_accessor(tree):
    """A column can be inserted with a default accessor."""

    tree.insert_column(1, "New Column")

    # The column was added
    assert_action_performed_with(
        tree,
        "insert column",
        index=1,
        heading="New Column",
        accessor="new_column",
    )
    assert tree.headings == ["Title", "New Column", "Value"]
    assert tree.accessors == ["key", "new_column", "value"]


def test_insert_column_no_headings(source):
    """A column can be inserted into a tree with no headings."""
    tree = toga.Tree(headings=None, accessors=["key", "value"], data=source)

    tree.insert_column(1, "New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        tree,
        "insert column",
        index=1,
        heading=None,
        accessor="extra",
    )
    assert tree.headings is None
    assert tree.accessors == ["key", "extra", "value"]


def test_insert_column_no_headings_missing_accessor(source):
    """An accessor is mandatory when adding a column to a tree with no headings."""
    tree = toga.Tree(headings=None, accessors=["key", "value"], data=source)

    with pytest.raises(
        ValueError,
        match=r"Must specify an accessor on a tree without headings",
    ):
        tree.insert_column(1, "New Column")


def test_append_column(tree):
    """A column can be appended."""
    tree.append_column("New Column", accessor="extra")

    # The column was added
    assert_action_performed_with(
        tree,
        "insert column",
        index=2,
        heading="New Column",
        accessor="extra",
    )
    assert tree.headings == ["Title", "Value", "New Column"]
    assert tree.accessors == ["key", "value", "extra"]


def test_remove_column_accessor(tree):
    """A column can be removed by accessor."""

    tree.remove_column("value")

    # The column was removed
    assert_action_performed_with(
        tree,
        "remove column",
        index=1,
    )
    assert tree.headings == ["Title"]
    assert tree.accessors == ["key"]


def test_remove_column_unknown_accessor(tree):
    """If the column named for removal doesn't exist, an error is raised."""
    with pytest.raises(ValueError, match=r"'unknown' is not in list"):
        tree.remove_column("unknown")


def test_remove_column_invalid_index(tree):
    """If the index specified doesn't exist, an error is raised."""
    with pytest.raises(IndexError, match=r"list assignment index out of range"):
        tree.remove_column(100)


def test_remove_column_index(tree):
    """A column can be removed by index."""

    tree.remove_column(1)

    # The column was removed
    assert_action_performed_with(
        tree,
        "remove column",
        index=1,
    )
    assert tree.headings == ["Title"]
    assert tree.accessors == ["key"]


def test_remove_column_negative_index(tree):
    """A column can be removed by index."""

    tree.remove_column(-2)

    # The column was removed
    assert_action_performed_with(
        tree,
        "remove column",
        index=0,
    )
    assert tree.headings == ["Value"]
    assert tree.accessors == ["value"]


def test_remove_column_no_headings(tree):
    """A column can be removed when there are no headings."""
    tree = toga.Tree(
        headings=None,
        accessors=["primus", "secondus"],
    )

    tree.remove_column(1)

    # The column was removed
    assert_action_performed_with(
        tree,
        "remove column",
        index=1,
    )
    assert tree.headings is None
    assert tree.accessors == ["primus"]


def test_deprecated_names(on_activate_handler):
    """Deprecated names still work."""

    # Can't specify both on_double_click and on_activate
    with pytest.raises(
        ValueError,
        match=r"Cannot specify both on_double_click and on_activate",
    ):
        toga.Tree(["First", "Second"], on_double_click=Mock(), on_activate=Mock())

    # on_double_click is redirected at construction
    with pytest.warns(
        DeprecationWarning,
        match="Tree.on_double_click has been renamed Tree.on_activate",
    ):
        tree = toga.Tree(["First", "Second"], on_double_click=on_activate_handler)

    # on_double_click accessor is redirected to on_activate
    with pytest.warns(
        DeprecationWarning,
        match="Tree.on_double_click has been renamed Tree.on_activate",
    ):
        assert tree.on_double_click._raw == on_activate_handler

    assert tree.on_activate._raw == on_activate_handler

    # on_double_click mutator is redirected to on_activate
    new_handler = Mock()
    with pytest.warns(
        DeprecationWarning,
        match="Tree.on_double_click has been renamed Tree.on_activate",
    ):
        tree.on_double_click = new_handler

    assert tree.on_activate._raw == new_handler



import importlib
import sys
from pathlib import Path

import toga
from toga.platform import get_platform_factory


class Paths:
    def __init__(self):
        self.factory = get_platform_factory()
        self._impl = self.factory.Paths(self)

    @property
    def toga(self) -> Path:
        """The path that contains the core Toga resources.

        This path should be considered read-only. You should not attempt to write
        files into this path.
        """
        return Path(toga.__file__).parent

    @property
    def app(self) -> Path:
        """The path of the folder that contains the definition of the app class.

        This path should be considered read-only. You should not attempt to write
        files into this path.
        """
        try:
            return Path(importlib.util.find_spec(toga.App.app.__module__).origin).parent
        except ValueError:
            # When running a single file `python path/to/myapp.py`, the app
            # won't have a module because it's the mainline. Default to the
            # path that contains the myapp.py.
            return Path(sys.argv[0]).absolute().parent

    @property
    def config(self) -> Path:
        """The platform-appropriate location for storing user configuration
        files associated with this app.
        """
        return self._impl.get_config_path()

    @property
    def data(self) -> Path:
        """The platform-appropriate location for storing user data associated
        with this app.
        """
        return self._impl.get_data_path()

    @property
    def cache(self) -> Path:
        """The platform-appropriate location for storing cache files associated
        with this app.

        It should be assumed that the operating system will purge the contents
        of this directory without warning if it needs to recover disk space.
        """
        return self._impl.get_cache_path()

    @property
    def logs(self) -> Path:
        """The platform-appropriate location for storing log files associated
        with this app.
        """
        return self._impl.get_logs_path()



from unittest.mock import Mock

import pytest

import toga
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
    attribute_value,
)
from toga_dummy.widgets.mapview import MapView as DummyMapView


@pytest.fixture
def pin_1():
    return toga.MapPin(toga.LatLng(10.0, 10.0), title="First", subtitle="thing")


@pytest.fixture
def pins(pin_1):
    return [
        pin_1,
        toga.MapPin((20.0, 20.0), title="Second"),
    ]


@pytest.fixture
def widget(pins):
    return toga.MapView(pins=pins)


def test_widget_created(widget):
    "A MapView can be created with minimal arguments"
    widget = toga.MapView()

    assert widget._impl.interface == widget
    assert_action_performed(widget, "create MapView")

    assert widget.location == toga.LatLng(-31.9559, 115.8606)
    assert widget.zoom == 11
    assert set(widget.pins) == set()
    assert repr(widget.pins) == "<MapPinSet (0 pins)>"
    assert widget._on_select._raw is None


def test_create_with_values(pins):
    """A MapView can be created with all available arguments"""
    on_select = Mock()

    widget = toga.MapView(location=(37.0, 42.0), zoom=4, pins=pins, on_select=on_select)

    assert widget._impl.interface == widget
    assert_action_performed(widget, "create MapView")

    assert widget.location == toga.LatLng(37.0, 42.0)
    assert widget.zoom == 4
    assert set(widget.pins) == set(pins)
    assert repr(widget.pins) == "<MapPinSet (2 pins)>"
    assert widget._on_select._raw == on_select


def test_latlng_properties():
    """LatLng objects behave like tuples."""
    # Create a LatLng object.
    pos = toga.LatLng(37.42, 42.37123456)

    # String representation is clean, and clipped to 6dp
    assert str(pos) == "(37.420000, 42.371235)"

    # Values can be accessed by attribute
    assert pos.lat == pytest.approx(37.42)
    assert pos.lng == pytest.approx(42.37123456)

    # Values can be accessed by position
    assert pos[0] == pytest.approx(37.42)
    assert pos[1] == pytest.approx(42.37123456)

    # LatLng can be compared with tuples
    assert pos == pytest.approx((37.42, 42.37123456))


@pytest.mark.parametrize(
    "title, subtitle, value",
    [
        ("The Title", None, "<MapPin @ (37.420000, 42.370000); The Title>"),
        (
            "The Title",
            "The Subtitle",
            "<MapPin @ (37.420000, 42.370000); The Title - The Subtitle>",
        ),
    ],
)
def test_pin_repr(title, subtitle, value):
    """The repr of a pin adapts to the properties that the pin has"""
    pin = toga.MapPin((37.42, 42.37), title=title, subtitle=subtitle)
    assert repr(pin) == value


def test_pin_location(widget):
    """Map pin location can be set and changed."""

    pin = toga.MapPin((37.42, 42.37), title="TheTitle", subtitle="TheSubtitle")

    assert isinstance(pin.location, toga.LatLng)
    assert pin.location == (37.42, 42.37)

    # Change the pin location before the pin is on the map
    EventLog.reset()

    # Pin Location can be changed with a tuple
    pin.location = (23.45, 67.89)
    assert isinstance(pin.location, toga.LatLng)
    assert pin.location == (23.45, 67.89)
    assert_action_not_performed(widget, "update pin")

    # Pin Location can be changed with a LatLng
    pin.location = toga.LatLng(12.34, 56.78)
    assert isinstance(pin.location, toga.LatLng)
    assert pin.location == (12.34, 56.78)
    assert_action_not_performed(widget, "update pin")

    # Add the pin to a map
    widget.pins.add(pin)
    assert_action_performed_with(widget, "add pin", pin=pin)

    # Pin Location can be changed while on the map
    pin.location = (23.45, 67.89)
    assert isinstance(pin.location, toga.LatLng)
    assert pin.location == (23.45, 67.89)
    assert_action_performed_with(widget, "update pin", pin=pin)

    # Remove the pin from the map
    EventLog.reset()
    widget.pins.remove(pin)
    assert_action_performed_with(widget, "remove pin", pin=pin)

    # Updating the location doesn't modify the map
    pin.location = toga.LatLng(12.34, 56.78)
    assert isinstance(pin.location, toga.LatLng)
    assert pin.location == (12.34, 56.78)
    assert_action_not_performed(widget, "update pin")


def test_pin_title(widget):
    """Map pin title can be set and changed."""
    pin = toga.MapPin((37.42, 42.37), title="The Title", subtitle="The Subtitle")
    assert pin.title == "The Title"

    # Change the pin title before the pin is on the map
    EventLog.reset()

    # Pin title can be changed to a string
    pin.title = "New Title"
    assert pin.title == "New Title"
    assert_action_not_performed(widget, "update pin")

    # If a non-string object is used, it is converted to a string
    pin.title = 12345
    assert pin.title == "12345"
    assert_action_not_performed(widget, "update pin")

    # Add the pin to a map
    widget.pins.add(pin)
    assert_action_performed_with(widget, "add pin", pin=pin)

    # Pin title can be changed while on the map
    pin.title = "On map"
    assert pin.title == "On map"
    assert_action_performed_with(widget, "update pin", pin=pin)

    # Remove the pin from the map
    EventLog.reset()
    widget.pins.remove(pin)
    assert_action_performed_with(widget, "remove pin", pin=pin)

    # Updating the title doesn't modify the map
    pin.title = "Off map"
    assert pin.title == "Off map"
    assert_action_not_performed(widget, "update pin")


def test_pin_subtitle(widget):
    """Map pin subtitle can be set and changed."""
    pin = toga.MapPin((37.42, 42.37), title="The Title", subtitle="The Subtitle")
    assert pin.subtitle == "The Subtitle"

    # Change the pin subtitle before the pin is on the map
    EventLog.reset()

    # Pin subtitle can be changed to a string
    pin.subtitle = "New Subtitle"
    assert pin.subtitle == "New Subtitle"
    assert_action_not_performed(widget, "update pin")

    # Pin subtitle can be set to None
    pin.subtitle = None
    assert pin.subtitle is None
    assert_action_not_performed(widget, "update pin")

    # If a non-string object is used, it is converted to a string
    pin.subtitle = 12345
    assert pin.subtitle == "12345"
    assert_action_not_performed(widget, "update pin")

    # Add the pin to a map
    widget.pins.add(pin)
    assert_action_performed_with(widget, "add pin", pin=pin)

    # Pin subtitle can be changed while on the map
    pin.subtitle = "On map"
    assert pin.subtitle == "On map"
    assert_action_performed_with(widget, "update pin", pin=pin)

    # Remove the pin from the map
    EventLog.reset()
    widget.pins.remove(pin)
    assert_action_performed_with(widget, "remove pin", pin=pin)

    # Updating the subtitle doesn't modify the map
    pin.subtitle = "Off map"
    assert pin.subtitle == "Off map"
    assert_action_not_performed(widget, "update pin")


def test_map_location(widget):
    """The map location can be changed."""

    # Location can be set with a tuple
    widget.location = (12.34, 56.78)

    assert isinstance(widget.location, toga.LatLng)
    assert widget.location == toga.LatLng(12.34, 56.78)

    # Location can be set with a LatLng
    widget.location = toga.LatLng(23.45, 67.89)

    assert isinstance(widget.location, toga.LatLng)
    assert widget.location == toga.LatLng(23.45, 67.89)


@pytest.mark.parametrize(
    "value, effective",
    [
        (-5, 0),  # Clipped to minimum
        (0, 0),  # minimum value
        (9, 9),  # mid range value
        (20, 20),  # maximum value
        (25, 20),  # clipped to maximum
        (3.14159, 3),  # converted to integer
        ("4", 4),  # converted to integer
    ],
)
def test_set_zoom(widget, value, effective):
    """Zoom value for the map can be set."""
    EventLog.reset()

    widget.zoom = value
    assert attribute_value(widget, "zoom") == effective
    # round trip the value
    assert widget.zoom == effective


@pytest.mark.parametrize(
    "value, exception",
    [
        (None, TypeError),
        ("a", ValueError),
    ],
)
def test_bad_zoom(widget, value, exception):
    """Bad values for zoom raise an error."""
    with pytest.raises(exception):
        widget.zoom = value


def test_add_remove_pin(widget):
    """Pins can be added and removed from the map."""
    pin = toga.MapPin((37.42, 42.37), title="The Title", subtitle="The Subtitle")

    # Initially 2 pins
    assert len(widget.pins) == 2

    # Add a pin
    widget.pins.add(pin)
    assert_action_performed_with(widget, "add pin", pin=pin)

    # There's now 3 pins
    assert len(widget.pins) == 3

    # Remove a pin
    widget.pins.remove(pin)
    assert_action_performed_with(widget, "remove pin", pin=pin)

    # There's now 2 pins
    assert len(widget.pins) == 2

    # Clear all pins
    widget.pins.clear()

    # There's now 0 pins
    assert len(widget.pins) == 0


def test_select_pin(widget, pin_1):
    """A pin on a map generates an event if selected."""
    # Set up a load handler
    on_select_handler = Mock()
    widget.on_select = on_select_handler

    # The load handler hasn't been called yet
    on_select_handler.assert_not_called()

    # Simulate a pin being selected
    widget._impl.simulate_pin_selected(pin_1)

    # handler has been invoked
    on_select_handler.assert_called_once_with(widget, pin=pin_1)


def test_disabled_on_select(widget):
    """If the backend doesn't support on_select, a warning is raised."""
    try:
        # Temporarily set the feature attribute on the backend
        DummyMapView.SUPPORTS_ON_SELECT = False

        # Instantiate a new widget with a hobbled backend.
        widget = toga.MapView()
        handler = Mock()

        # Setting the handler raises a warning
        with pytest.warns(
            toga.NotImplementedWarning,
            match=r"\[Dummy\] Not implemented: MapView\.on_select",
        ):
            widget.on_select = handler

        # But the handler is still installed
        assert widget.on_select._raw == handler
    finally:
        # Clear the feature attribute.
        del DummyMapView.SUPPORTS_ON_SELECT



from travertino.size import at_least

from toga.style.pack import COLUMN, ROW, Pack

from ..utils import ExampleNode, ExampleViewport, assert_layout


def test_row_expanding_intrinsic():
    """Children in a row layout with fixed size don't expand, even if their hints allow
    it."""
    root = ExampleNode(
        "app",
        style=Pack(direction=ROW),
        children=[
            ExampleNode(
                "first",
                style=Pack(width=100, height=100),
                size=(at_least(0), at_least(0)),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=ROW, width=100, height=100),
                size=(at_least(50), at_least(50)),
                children=[
                    ExampleNode(
                        "child",
                        style=Pack(width=50, height=50),
                        size=(at_least(0), at_least(0)),
                    ),
                ],
            ),
            ExampleNode(
                "third",
                style=Pack(width=0, height=0),  # Explicitly 0 sized
                size=(at_least(0), at_least(0)),
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (200, 100),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (100, 100),
                },
                {
                    "origin": (100, 0),
                    "content": (100, 100),
                    "children": [
                        {"origin": (100, 0), "content": (50, 50)},
                    ],
                },
                {
                    "origin": (200, 0),
                    "content": (0, 0),
                },
            ],
        },
    )


def test_row_fixed_intrinsic():
    """Children in a row layout without an explicit size, but a fixed intrinsic width,
    don't expand."""
    root = ExampleNode(
        "app",
        style=Pack(direction=ROW),
        children=[
            ExampleNode(
                "first",
                style=Pack(),
                size=(100, 100),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=ROW),
                children=[
                    ExampleNode(
                        "child",
                        style=Pack(),
                        size=(50, 50),
                    ),
                ],
            ),
            ExampleNode(  # Explicitly 0 sized
                "third",
                style=Pack(),
                size=(0, 0),
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (150, 100),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (100, 100),
                },
                {
                    "origin": (100, 0),
                    "content": (50, 480),
                    "children": [
                        {"origin": (100, 0), "content": (50, 50)},
                    ],
                },
                {
                    "origin": (150, 0),
                    "content": (0, 0),
                },
            ],
        },
    )


def test_column_expanding_intrinsic():
    """Children in a column layout with fixed size don't expand, even if their hints
    allow it."""
    root = ExampleNode(
        "app",
        style=Pack(direction=COLUMN),
        children=[
            ExampleNode(
                "first",
                style=Pack(width=100, height=100),
                size=(at_least(0), at_least(0)),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=COLUMN, width=100, height=100),
                size=(at_least(50), at_least(50)),
                children=[
                    ExampleNode(
                        "child",
                        style=Pack(width=50, height=50),
                        size=(at_least(0), at_least(0)),
                    ),
                ],
            ),
            ExampleNode(  # Explicitly 0 sized
                "third",
                style=Pack(width=0, height=0),
                size=(at_least(0), at_least(0)),
            ),
        ],
    )

    # Normal size
    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (100, 200),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {"origin": (0, 0), "content": (100, 100)},
                {
                    "origin": (0, 100),
                    "content": (100, 100),
                    "children": [
                        {"origin": (0, 100), "content": (50, 50)},
                    ],
                },
                {
                    "origin": (0, 200),
                    "content": (0, 0),
                },
            ],
        },
    )


def test_column_fixed_intrinsic():
    """Children in a column layout without an explicit size, but a fixed intrinsic
    width, don't expand."""
    root = ExampleNode(
        "app",
        style=Pack(direction=COLUMN),
        children=[
            ExampleNode(
                "first",
                style=Pack(),
                size=(100, 100),
            ),
            ExampleNode(
                "second",
                style=Pack(direction=COLUMN),
                children=[
                    ExampleNode(
                        "child",
                        style=Pack(),
                        size=(50, 50),
                    ),
                ],
            ),
            ExampleNode(  # Explicitly 0 sized
                "third",
                style=Pack(),
                size=(0, 0),
            ),
        ],
    )

    root.style.layout(root, ExampleViewport(640, 480))
    assert_layout(
        root,
        (100, 150),
        (640, 480),
        {
            "origin": (0, 0),
            "content": (640, 480),
            "children": [
                {
                    "origin": (0, 0),
                    "content": (100, 100),
                },
                {
                    "origin": (0, 100),
                    "content": (640, 50),
                    "children": [
                        {"origin": (0, 100), "content": (50, 50)},
                    ],
                },
                {
                    "origin": (0, 150),
                    "content": (0, 0),
                },
            ],
        },
    )



from __future__ import annotations

from typing import Literal

from .base import Widget


class ActivityIndicator(Widget):
    def __init__(
        self,
        id=None,
        style=None,
        running: bool = False,
    ):
        """Create a new ActivityIndicator widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        :param running: Describes whether the indicator is running at the
            time it is created.
        """
        super().__init__(id=id, style=style)

        self._impl = self.factory.ActivityIndicator(interface=self)

        if running:
            self.start()

    @property
    def enabled(self) -> Literal[True]:
        """Is the widget currently enabled? i.e., can the user interact with the widget?

        ActivityIndicator widgets cannot be disabled; this property will always return
        True; any attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value: bool) -> None:
        pass

    def focus(self) -> None:
        "No-op; ActivityIndicator cannot accept input focus"
        pass

    @property
    def is_running(self) -> bool:
        """Determine if the activity indicator is currently running.

        Use ``start()`` and ``stop()`` to change the running state.

        True if this activity indicator is running; False otherwise.
        """
        return self._impl.is_running()

    def start(self) -> None:
        """Start the activity indicator.

        If the activity indicator is already started, this is a no-op.
        """
        if not self.is_running:
            self._impl.start()

    def stop(self) -> None:
        """Stop the activity indicator.

        If the activity indicator is already stopped, this is a no-op.
        """
        if self.is_running:
            self._impl.stop()



from __future__ import annotations

# Use the Travertino font definitions as-is
from travertino import constants
from travertino.constants import (
    BOLD,
    CURSIVE,
    FANTASY,
    ITALIC,
    MESSAGE,
    MONOSPACE,
    NORMAL,
    OBLIQUE,
    SANS_SERIF,
    SERIF,
    SMALL_CAPS,
    SYSTEM,
)
from travertino.fonts import Font as BaseFont

import toga
from toga.platform import get_platform_factory

SYSTEM_DEFAULT_FONTS = {SYSTEM, MESSAGE, SERIF, SANS_SERIF, CURSIVE, FANTASY, MONOSPACE}
SYSTEM_DEFAULT_FONT_SIZE = -1
FONT_WEIGHTS = {NORMAL, BOLD}
FONT_STYLES = {NORMAL, ITALIC, OBLIQUE}
FONT_VARIANTS = {NORMAL, SMALL_CAPS}

_REGISTERED_FONT_CACHE = {}


class Font(BaseFont):
    def __init__(
        self,
        family: str,
        size: int | str,
        *,
        weight: str = NORMAL,
        style: str = NORMAL,
        variant: str = NORMAL,
    ):
        """Constructs a reference to a font.

        This class should be used when an API requires an explicit font reference (e.g.
        :any:`Context.write_text`). In all other cases, fonts in Toga are controlled
        using the style properties linked below.

        :param family: The :ref:`font family <pack-font-family>`.
        :param size: The :ref:`font size <pack-font-size>`.
        :param weight: The :ref:`font weight <pack-font-weight>`.
        :param style: The :ref:`font style <pack-font-style>`.
        :param variant: The :ref:`font variant <pack-font-variant>`.
        """
        super().__init__(family, size, weight=weight, style=style, variant=variant)
        self.factory = get_platform_factory()
        self._impl = self.factory.Font(self)

    def __str__(self) -> str:
        size = (
            "default size"
            if self.size == SYSTEM_DEFAULT_FONT_SIZE
            else f"{self.size}pt"
        )
        weight = f" {self.weight}" if self.weight != NORMAL else ""
        variant = f" {self.variant}" if self.variant != NORMAL else ""
        style = f" {self.style}" if self.style != NORMAL else ""
        return f"{self.family} {size}{weight}{variant}{style}"

    @staticmethod
    def register(family, path, *, weight=NORMAL, style=NORMAL, variant=NORMAL):
        """Registers a file-based font.

        **Note:** This is not currently supported on macOS or iOS.

        :param family: The :ref:`font family <pack-font-family>`.
        :param path: The path to the font file. This can be an absolute path, or a path
            relative to the module that defines your :any:`App` class.
        :param weight: The :ref:`font weight <pack-font-weight>`.
        :param style: The :ref:`font style <pack-font-style>`.
        :param variant: The :ref:`font variant <pack-font-variant>`.
        """
        font_key = Font._registered_font_key(family, weight, style, variant)
        _REGISTERED_FONT_CACHE[font_key] = str(toga.App.app.paths.app / path)

    @staticmethod
    def _registered_font_key(family, weight, style, variant):
        if weight not in constants.FONT_WEIGHTS:
            weight = NORMAL
        if style not in constants.FONT_STYLES:
            style = NORMAL
        if variant not in constants.FONT_VARIANTS:
            variant = NORMAL

        return family, weight, style, variant



from unittest.mock import Mock

import pytest
from pytest import approx, fixture, raises

import toga
from toga_dummy.utils import assert_action_performed, attribute_value

INITIAL_VALUE = 50
INITIAL_MIN = 0
INITIAL_MAX = 100
INITIAL_TICK_VALUE = 6
INITIAL_TICK_COUNT = 11
INITIAL_ENABLED = True


@fixture
def on_change():
    return Mock()


@fixture
def slider(on_change):
    return toga.Slider(
        value=INITIAL_VALUE,
        min=INITIAL_MIN,
        max=INITIAL_MAX,
        on_change=on_change,
        enabled=INITIAL_ENABLED,
        tick_count=INITIAL_TICK_COUNT,
    )


def test_widget_created(slider, on_change):
    assert slider._impl.interface == slider
    assert_action_performed(slider, "create Slider")


@pytest.mark.parametrize(
    "value",
    [
        INITIAL_VALUE,
        30,  # int
        30.0,  # float
        30.5,  # non-integral float
    ],
)
def test_set_value(slider, on_change, value):
    slider.tick_count = None
    slider.value = value
    assert_value(
        slider,
        on_change,
        value,
        change_count=0 if (value == INITIAL_VALUE) else 1,
    )

    # Resetting the same value should not call on_change.
    on_change.reset_mock()
    slider.value = value
    assert_value(slider, on_change, value, change_count=0)


def test_set_value_to_be_min(slider, on_change):
    slider.value = INITIAL_MIN
    assert_value(slider, on_change, value=INITIAL_MIN, tick_value=1, change_count=1)


def test_set_value_to_be_max(slider, on_change):
    slider.value = INITIAL_MAX
    assert_value(
        slider,
        on_change,
        value=INITIAL_MAX,
        tick_value=INITIAL_TICK_COUNT,
        change_count=1,
    )


def test_set_value_to_be_too_small(slider, on_change):
    """Setting the value below the minimum results in clipping."""
    slider.value = INITIAL_MIN - 1
    assert_value(slider, on_change, tick_value=1, value=INITIAL_MIN, change_count=1)


def test_set_value_to_be_too_big(slider, on_change):
    """Setting the value above the maximum results in clipping."""
    slider.value = INITIAL_MAX + 1
    assert_value(slider, on_change, tick_value=11, value=INITIAL_MAX, change_count=1)


def test_set_tick_value_between_min_and_max(slider, on_change):
    value = 30
    tick_value = 4
    slider.tick_value = tick_value
    assert_value(slider, on_change, value=value, tick_value=tick_value, change_count=1)


def test_set_tick_value_to_be_min(slider, on_change):
    slider.tick_value = 1
    assert_value(slider, on_change, value=INITIAL_MIN, tick_value=1, change_count=1)


def test_set_tick_value_to_be_max(slider, on_change):
    slider.tick_value = INITIAL_TICK_COUNT
    assert_value(
        slider,
        on_change,
        value=INITIAL_MAX,
        tick_value=INITIAL_TICK_COUNT,
        change_count=1,
    )


def test_set_tick_value_to_be_too_small(slider, on_change):
    """Setting the tick value to less than the min results in clipping."""
    slider.tick_value = 0
    assert_value(slider, on_change, tick_value=1, value=INITIAL_MIN, change_count=1)


def test_set_tick_value_to_be_too_big(slider, on_change):
    """Setting the tick value to greater than the max results in clipping."""
    slider.tick_value = INITIAL_TICK_COUNT + 1
    assert_value(slider, on_change, tick_value=11, value=INITIAL_MAX, change_count=1)


def test_tick_value_without_tick_count(slider, on_change):
    slider.tick_count = None
    with raises(ValueError, match="cannot set tick value when tick count is None"):
        slider.tick_value = 4


def test_set_tick_value_to_none(slider, on_change):
    slider.tick_count = None
    slider.tick_value = None
    assert_value(slider, on_change, INITIAL_VALUE, tick_value=None)


def test_set_tick_value_to_none_with_tick_count(slider, on_change):
    with raises(
        ValueError, match="cannot set tick value to None when tick count is not None"
    ):
        slider.tick_value = None


def test_increasing_by_value(slider, on_change):
    """Slider.value supports += syntax."""
    delta = 20
    tick_delta = 2
    slider.value += delta
    assert_value(
        slider,
        on_change,
        value=INITIAL_VALUE + delta,
        tick_value=INITIAL_TICK_VALUE + tick_delta,
        change_count=1,
    )


def test_decreasing_by_value(slider, on_change):
    """Slider.value supports -= syntax."""
    delta = 20
    tick_delta = 2
    slider.value -= delta
    assert_value(
        slider,
        on_change,
        value=INITIAL_VALUE - delta,
        tick_value=INITIAL_TICK_VALUE - tick_delta,
        change_count=1,
    )


def test_increasing_by_ticks(slider, on_change):
    """Slider.tick_value supports += syntax."""
    delta = 20
    tick_delta = 2
    slider.tick_value += tick_delta
    assert_value(
        slider,
        on_change,
        value=INITIAL_VALUE + delta,
        tick_value=INITIAL_TICK_VALUE + tick_delta,
        change_count=1,
    )


def test_decreasing_by_ticks(slider, on_change):
    """Slider.tick_value supports -= syntax."""
    delta = 20
    tick_delta = 2
    slider.tick_value -= tick_delta
    assert_value(
        slider,
        on_change,
        value=INITIAL_VALUE - delta,
        tick_value=INITIAL_TICK_VALUE - tick_delta,
        change_count=1,
    )


@pytest.mark.parametrize(
    "min, max, value",
    [
        (0, 100, 50),  # Unchanged
        (30, 90, 50),  # Reduced, but still contains current value
        (50, 100, 50),  # Barely contains current value
        (0, 50, 50),
        (60, 100, 60),  # Does not contain current value
        (0, 30, 30),
    ],
)
def test_range(slider, on_change, min, max, value):
    """Setting the range clamps the existing value."""
    slider.tick_count = None
    slider.min = min
    slider.max = max
    assert isinstance(slider.min, float)
    assert slider.min == pytest.approx(min)
    assert isinstance(slider.max, float)
    assert slider.max == pytest.approx(max)

    assert_value(
        slider,
        on_change,
        value,
        change_count=0 if (value == INITIAL_VALUE) else 1,
    )


@pytest.mark.parametrize(
    "new_min, new_value, new_max",
    [
        [-5, 5, 10],  # less than old min
        [5, 5, 10],  # more than old min
        [6, 6, 10],  # more than old min and value
        [15, 15, 15],  # more than old min, value and max
    ],
)
def test_min_clipping(slider, new_min, new_value, new_max):
    slider.tick_count = None
    slider.min = 0
    slider.value = 5
    slider.max = 10

    slider.min = new_min
    assert slider.min == new_min
    assert slider.value == new_value
    assert slider.max == new_max


@pytest.mark.parametrize(
    "new_min, new_value, new_max",
    [
        [0, 5, 15],  # more than old max
        [0, 5, 5],  # less than old max
        [0, 4, 4],  # less than old max and value
        [-5, -5, -5],  # less than old max, value and min
    ],
)
def test_max_clipping(slider, new_min, new_value, new_max):
    slider.tick_count = None
    slider.min = 0
    slider.value = 5
    slider.max = 10

    slider.max = new_max
    assert slider.min == new_min
    assert slider.value == new_value
    assert slider.max == new_max


def test_set_enabled_with_working_values(slider, on_change):
    assert slider.enabled == INITIAL_ENABLED
    slider.enabled = False
    assert slider.enabled is False


def test_get_tick_count(slider, on_change):
    tick_count = slider.tick_count
    assert INITIAL_TICK_COUNT == tick_count


TICK_RANGE = (10, 110)
TICK_PARAM_NAMES = "tick_count, tick_step, tick_value, value"
TICK_PARAM_VALUES = [
    (None, None, None, 50),
    (11, 10, 5, 50),
    (5, 25, 3, 60),  # Rounded up
    (4, 33.33333, 2, 43.33333),  # Rounded down
    (2, 100, 1, 10),  # Rounded down to the minimum
]


@pytest.mark.parametrize(TICK_PARAM_NAMES, TICK_PARAM_VALUES)
def test_set_tick_count(slider, on_change, tick_count, tick_step, tick_value, value):
    """Setting the tick count rounds the existing value to the nearest tick."""
    slider.min = TICK_RANGE[0]
    slider.max = TICK_RANGE[1]
    # Setting min and max will send change signals.
    on_change.reset_mock()

    slider.tick_count = tick_count
    assert slider.tick_count == tick_count
    assert attribute_value(slider, "tick_count") == tick_count

    if tick_step is None:
        assert slider.tick_step is None
    else:
        assert isinstance(slider.tick_step, float)
        assert slider.tick_step == approx(tick_step)

    assert_value(
        slider,
        on_change,
        approx(value),
        tick_value=tick_value,
        change_count=0 if (value == INITIAL_VALUE) else 1,
    )


@pytest.mark.parametrize(TICK_PARAM_NAMES, TICK_PARAM_VALUES)
def test_set_value_with_tick_count(
    slider, on_change, tick_count, tick_step, tick_value, value
):
    """Setting the value rounds it to the nearest tick."""
    slider.min = TICK_RANGE[0]
    slider.max = TICK_RANGE[1]
    # Setting min and max will send change signals.
    on_change.reset_mock()

    slider.tick_count = tick_count
    slider.value = TICK_RANGE[1]

    on_change.reset_mock()
    slider.value = INITIAL_VALUE
    assert_value(
        slider, on_change, approx(value), tick_value=tick_value, change_count=1
    )

    # Resetting the same value should round to the same result, so on_change should not be
    # called.
    on_change.reset_mock()
    slider.value = INITIAL_VALUE
    assert_value(
        slider, on_change, approx(value), tick_value=tick_value, change_count=0
    )


def test_set_tick_count_too_small(slider, on_change):
    for tick_count in [1, 0, -1]:
        with raises(ValueError, match="tick count must be at least 2"):
            slider.tick_count = tick_count


def test_focus(slider, on_change):
    slider.focus()
    assert_action_performed(slider, "focus")


def test_set_on_change():
    on_change = Mock()
    slider = toga.Slider(on_change=on_change)
    assert slider.on_change._raw == on_change


def test_set_on_press():
    on_press = Mock()
    slider = toga.Slider(on_press=on_press)
    assert slider.on_press._raw == on_press


def test_set_on_release():
    on_release = Mock()
    slider = toga.Slider(on_release=on_release)
    assert slider.on_release._raw == on_release


def assert_value(slider, on_change, value, *, tick_value=None, change_count=0):
    """Asserts that the slider's `value` and `tick_value` attributes have the given
    values, and that `on_change` has been called `change_count` times."""
    assert isinstance(slider.tick_value, (int, type(None)))
    assert slider.tick_value == tick_value

    assert isinstance(slider.value, float)
    assert slider.value == value
    assert attribute_value(slider, "value") == value

    assert on_change.call_count == change_count


class DummyIntImpl(toga.widgets.slider.IntSliderImpl):
    def __init__(self):
        super().__init__()
        self.interface = Mock()

    def get_int_value(self):
        return self.int_value

    def set_int_value(self, value):
        self.int_value = value

    def get_int_max(self):
        return self.int_max

    def set_int_max(self, max):
        self.int_max = max

    def set_ticks_visible(self, visible):
        self.ticks_visible = visible


def test_int_impl_continuous():
    impl = DummyIntImpl()
    assert impl.get_value() == 0

    impl.set_min(0)
    assert impl.get_min() == 0
    impl.set_max(1)
    assert impl.get_max() == 1
    impl.set_tick_count(None)
    assert impl.get_tick_count() is None
    assert impl.int_max == 10000
    assert impl.ticks_visible is False

    # Values should be converted into ints on a scale of 0 to 10000.
    for value, int_value in [
        (0, 0),
        (0.5, 5000),
        (1, 10000),
        (0.00004, 0),
        (0.00006, 1),
        (0.0001, 1),
    ]:
        # At this level, the value should be round-tripped, not rounded.
        impl.set_value(value)
        assert impl.int_value == int_value
        assert impl.get_value() == value

    # Range that doesn't start at zero
    impl.set_min(-0.4)
    assert impl.get_min() == pytest.approx(-0.4)
    impl.set_max(0.6)
    assert impl.get_max() == pytest.approx(0.6)
    impl.set_value(0.5)
    assert impl.get_value() == 0.5
    assert impl.int_value == 9000
    assert impl.int_max == 10000

    # Empty range
    impl.set_min(0)
    impl.set_max(0)
    impl.set_value(0)
    assert impl.get_value() == 0
    assert impl.int_value == 0
    assert impl.int_max == 10000

    # Empty range that doesn't start at zero
    impl.set_min(1)
    impl.set_max(1)
    impl.set_value(1)
    assert impl.get_value() == 1
    assert impl.int_value == 0
    assert impl.int_max == 10000


def test_int_impl_discrete():
    impl = DummyIntImpl()
    assert impl.get_value() == 0

    impl.set_min(0)
    assert impl.get_min() == 0
    impl.set_max(1)
    assert impl.get_max() == 1
    impl.set_tick_count(9)
    assert impl.get_tick_count() == 9
    assert impl.int_max == 8
    assert impl.ticks_visible is True

    # Values should be converted into ints on a scale of one int per tick, rounding to
    # the nearest one.
    for value, int_value in [
        (0, 0),  # 0.000
        (1, 8),  # 1.000
        (0.1, 1),  # 0.125
        (0.2, 2),  # 0.250
        (0.3, 2),  # 0.250
        (0.4, 3),  # 0.375
    ]:
        # At this level, the value should be round-tripped, not rounded.
        impl.set_value(value)
        assert impl.get_value() == value
        assert impl.int_value == int_value

    # Range that doesn't start at zero
    impl.set_min(-0.4)
    assert impl.get_min() == pytest.approx(-0.4)
    impl.set_max(0.6)
    assert impl.get_max() == pytest.approx(0.6)
    impl.set_value(0.5)
    assert impl.get_value() == 0.5
    assert impl.int_value == 7
    assert impl.int_max == 8

    # Empty range
    impl.set_min(0)
    impl.set_max(0)
    impl.set_value(0)
    assert impl.get_value() == 0
    assert impl.int_value == 0
    assert impl.int_max == 8

    # Empty range that doesn't start at zero
    impl.set_min(1)
    impl.set_max(1)
    impl.set_value(1)
    assert impl.get_value() == 1
    assert impl.int_value == 0
    assert impl.int_max == 8


@pytest.mark.parametrize(
    "tick_count, data",
    [
        (
            None,
            [(0, 0), (1, 10000), (0.0001, 1), (0.5, 5000)],
        ),
        (
            9,
            [(0, 0), (1, 8), (0.125, 1), (0.250, 2)],
        ),
    ],
)
def test_int_impl_on_change(tick_count, data):
    """Ints should be converted into values correctly."""
    impl = DummyIntImpl()
    impl.set_min(0)
    impl.set_max(1)
    impl.set_tick_count(tick_count)
    for value, int_value in data:
        impl.interface.reset_mock()
        impl.int_value = int_value
        impl.on_change()
        assert impl.get_value() == approx(value)
        impl.interface.on_change.assert_called_once_with()


def test_deprecated():
    """Check the deprecated min/max naming."""
    # Can't specify min and range
    with pytest.raises(
        ValueError,
        match=r"range cannot be specified if min and max are specified",
    ):
        toga.Slider(min=2, range=(2, 4))

    # Can't specify max and range
    with pytest.raises(
        ValueError,
        match=r"range cannot be specified if min and max are specified",
    ):
        toga.Slider(max=4, range=(2, 4))

    # Can't specify min and max and range
    with pytest.raises(
        ValueError,
        match=r"range cannot be specified if min and max are specified",
    ):
        toga.Slider(min=2, max=4, range=(2, 4))

    # Range is deprecated
    with pytest.warns(
        DeprecationWarning,
        match="Slider.range has been deprecated in favor of Slider.min and Slider.max",
    ):
        widget = toga.Slider(range=(2, 4))

    # range is converted to min/max
    assert widget.min == pytest.approx(2)
    assert widget.max == pytest.approx(4)

    with pytest.warns(
        DeprecationWarning,
        match="Slider.range has been deprecated in favor of Slider.min and Slider.max",
    ):
        assert widget.range == (pytest.approx(2), pytest.approx(4))

    # range is converted to min/max
    with pytest.warns(
        DeprecationWarning,
        match="Slider.range has been deprecated in favor of Slider.min and Slider.max",
    ):
        widget.range = (6, 8)

    assert widget.min == pytest.approx(6)
    assert widget.max == pytest.approx(8)



from decimal import Decimal
from unittest.mock import Mock

import pytest

import toga
from toga.widgets.numberinput import _clean_decimal, _clean_decimal_str
from toga_dummy.utils import (
    EventLog,
    assert_action_performed,
    attribute_value,
)


@pytest.fixture
def widget():
    return toga.NumberInput(step="0.01")


def test_widget_created():
    """A NumberInput can be created with minimal arguments."""
    widget = toga.NumberInput()

    assert widget._impl.interface == widget
    assert_action_performed(widget, "create NumberInput")

    assert not widget.readonly
    assert widget.value is None
    assert widget.step == Decimal("1")
    assert widget.min is None
    assert widget.max is None
    assert widget._on_change._raw is None


def test_create_with_values():
    """A NumberInput can be created with initial values."""
    on_change = Mock()

    widget = toga.NumberInput(
        value=Decimal("2.71828"),
        step=0.001,
        min=-42,
        max=420,
        readonly=True,
        on_change=on_change,
    )
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create NumberInput")

    assert widget.readonly
    assert widget.value == Decimal("2.718")
    assert widget.step == Decimal("0.001")
    assert widget.min == Decimal("-42")
    assert widget.max == Decimal("420")
    assert widget._on_change._raw == on_change

    # Change handler hasn't been invoked
    on_change.assert_not_called()


@pytest.mark.parametrize(
    "value, expected",
    [
        # float
        (1.23, Decimal("1.23")),
        (1.23e4, Decimal("12300")),
        # Integer
        (123, Decimal("123")),
        # Float and integer specified as string
        ("1.23", Decimal("1.23")),
        ("123", Decimal("123")),
        ("1.23e+4", Decimal("1.23e+4")),
        # Excessive precision is rounded to same number of decimal places as step
        # (see QUANTIZE_PARAMS for related tests).
        ("1.23456", Decimal("1.23")),
        ("1.235", Decimal("1.24")),
        # Decimal
        (Decimal("1.23"), Decimal("1.23")),
        # Empty string
        ("", None),
        # None
        (None, None),
    ],
)
def test_value(widget, value, expected):
    """The value of the widget can be set."""
    # Clear the event log and validator mock
    EventLog.reset()

    # Define and set a new change callback
    on_change_handler = Mock()
    widget.on_change = on_change_handler

    widget.value = value
    assert widget.value == expected

    # test backend has the right value
    assert attribute_value(widget, "value") == expected

    # A refresh was performed
    assert_action_performed(widget, "refresh")

    # change handler was invoked
    on_change_handler.assert_called_once_with(widget)


@pytest.mark.parametrize(
    "value",
    [
        object(),  # Not convertible
        "Not a number",  # Non-numerical string
    ],
)
def test_bad_value(widget, value):
    """If a value can't be converted into a decimal, an error is raised."""
    with pytest.raises(ValueError, match=r"value must be a number or None"):
        widget.value = value


@pytest.mark.parametrize(
    "value, expected",
    [
        # float
        (1.23, Decimal("1.23")),
        (1.23e4, Decimal("12300")),
        # Integer
        (123, Decimal("123")),
        # Float and integer specified as string
        ("1.23", Decimal("1.23")),
        ("123", Decimal("123")),
        ("1.23e+4", Decimal("1.23e+4")),
        # Decimal
        (Decimal("1.23"), Decimal("1.23")),
    ],
)
def test_step(widget, value, expected):
    """The step of the widget can be set."""
    widget.step = value
    assert widget.step == expected

    # test backend has the right value
    assert attribute_value(widget, "step") == expected


QUANTIZE_PARAMS = (
    "step, expected",
    [
        ("0.0001", Decimal("12.3456")),
        ("0.001", Decimal("12.346")),
        ("0.009", Decimal("12.346")),
        ("0.010", Decimal("12.346")),
        ("0.01", Decimal("12.35")),
        ("0.1", Decimal("12.3")),
        ("1", Decimal("12")),
        ("10", Decimal("12")),
    ],
)


@pytest.mark.parametrize(*QUANTIZE_PARAMS)
def test_quantization(widget, step, expected):
    """The value is quantized to the precision of the step."""
    widget.step = step
    widget.value = 12.3456

    # The value has been quantized to the step
    assert widget.value == expected


@pytest.mark.parametrize(*QUANTIZE_PARAMS)
def test_quantize_on_retrieval(widget, step, expected):
    """A widget's value will be quantized on retrieval."""
    widget.step = step

    # Inject a raw attribute value.
    widget._impl._set_value("value", 12.3456)
    assert widget.value == expected


@pytest.mark.parametrize(
    "value",
    [
        object(),  # Not convertible
        "Not a number",  # Non-numerical string
        "",  # Empty string
        None,
    ],
)
def test_bad_step(widget, value):
    """If a step can't be converted into a decimal, an error is raised."""
    with pytest.raises(ValueError, match=r"step must be a number"):
        widget.step = "not a number"


@pytest.mark.parametrize(
    "value, expected",
    [
        # float; approximate because of float conversion
        (1.23, Decimal("1.23")),
        (1.23e4, Decimal("12300")),
        # Integer
        (123, Decimal("123")),
        # Integer (but a value that evaluates as false)
        (0, Decimal("0")),
        # Float and integer specified as string
        ("1.23", Decimal("1.23")),
        ("123", Decimal("123")),
        ("1.23e+4", Decimal("1.23e+4")),
        # Excessive precision is truncated to step value
        ("1.23456", Decimal("1.23")),
        # Decimal
        (Decimal("1.23"), Decimal("1.23")),
        # Empty string
        ("", None),
        # None
        (None, None),
    ],
)
def test_min(widget, value, expected):
    """The min of the widget can be set."""
    widget.min = value
    assert widget.min == expected

    # test backend has the right value
    assert attribute_value(widget, "min") == expected


@pytest.mark.parametrize(
    "value",
    [
        object(),  # Not convertible
        "Not a number",  # Non-numerical string
    ],
)
def test_bad_min(widget, value):
    """If a min can't be converted into a decimal, an error is raised."""
    with pytest.raises(ValueError, match=r"min must be a number or None"):
        widget.min = value


def test_min_greater_than_max(widget):
    """If the new min value exceeds the max value, the max value is clipped."""
    widget.max = 10
    widget.min = 100

    assert widget.max == 100
    assert widget.min == 100


@pytest.mark.parametrize(*QUANTIZE_PARAMS)
def test_min_quantized(widget, step, expected):
    """An existing min value is re-quantized after a change in step."""
    # Set a small step so that the min value isn't quantized
    widget.step = 0.00000001
    widget.min = 12.3456

    # Set a new minimum
    widget.step = step

    # The minimum has been re-quantized
    assert widget.min == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        # float
        (1.23, Decimal("1.23")),
        (1.23e4, Decimal("12300")),
        # Integer
        (123, Decimal("123")),
        # Integer (but a value that evaluates as false)
        (0, Decimal("0")),
        # Float and integer specified as string
        ("1.23", Decimal("1.23")),
        ("123", Decimal("123")),
        ("1.23e+4", Decimal("1.23e+4")),
        # Excessive precision is truncated to step value
        ("1.23456", Decimal("1.23")),
        # Decimal
        (Decimal("1.23"), Decimal("1.23")),
        # Empty string
        ("", None),
        # None
        (None, None),
    ],
)
def test_max(widget, value, expected):
    """The max of the widget can be set."""
    widget.max = value
    assert widget.max == expected

    # test backend has the right value
    assert attribute_value(widget, "max") == expected


@pytest.mark.parametrize(
    "value",
    [
        object(),  # Not convertible
        "Not a number",  # Non-numerical string
    ],
)
def test_bad_max(widget, value):
    """If a max can't be converted into a decimal, an error is raised."""
    with pytest.raises(ValueError, match=r"max must be a number or None"):
        widget.max = value


def test_max_less_than_min(widget):
    """If the new max value is less than the min value, the min value is clipped."""
    widget.min = 100
    widget.max = 10

    assert widget.max == 10
    assert widget.min == 10


@pytest.mark.parametrize(*QUANTIZE_PARAMS)
def test_max_quantized(widget, step, expected):
    """An existing max value is re-quantized after a change in step."""
    # Set a small step so that the max value isn't quantized
    widget.step = 0.00000001
    widget.max = 12.3456

    # Set a new maximum
    widget.step = step

    # The maximum has been re-quantized
    assert widget.max == expected


@pytest.mark.parametrize(
    "min, max, provided, clipped",
    [
        (10, 20, 15, Decimal(15)),
        (10, 20, 25, Decimal(20)),
        (10, 20, 5, Decimal(10)),
        (10, 20, None, None),
        # Check min/max values of 0
        (0, 20, -15, Decimal(0)),
        (-10, 0, 25, Decimal(0)),
    ],
)
def test_clip_on_value_change(widget, min, max, provided, clipped):
    """A widget's value will be clipped inside the min/max range."""
    widget.min = min
    widget.max = max

    widget.value = provided
    assert widget.value == clipped


@pytest.mark.parametrize(
    "min, max, provided, clipped",
    [
        (10, 20, 15, Decimal(15)),
        (10, 20, 25, Decimal(20)),
        (10, 20, 5, Decimal(10)),
        (10, 20, None, None),
        # Check min/max values of 0
        (0, 20, -15, Decimal(0)),
        (-10, 0, 25, Decimal(0)),
        # Check a "raw" value of 0
        (10, 20, 0, Decimal(10)),
        (-20, -10, 0, Decimal(-10)),
    ],
)
def test_clip_on_retrieval(widget, min, max, provided, clipped):
    """A widget's value will be clipped if the widget has a value outside the min/max
    range."""
    widget.min = min
    widget.max = max

    # Inject a raw attribute value.
    widget._impl._set_value("value", provided)
    assert widget.value == clipped


@pytest.mark.parametrize(
    "value, new_max, clipped",
    [
        (10, 15, Decimal(10)),
        (10, 5, Decimal(5)),
        (10, None, Decimal(10)),
        (None, None, None),
        (None, 5, None),
    ],
)
def test_clip_on_max_change(widget, value, new_max, clipped):
    """A widget's value will be clipped if the max value changes."""
    # Set an initial max, and a value that is less than it.
    widget.max = 20
    widget.value = value

    # Set a new max
    widget.max = new_max
    # Value might be clipped
    assert widget.value == clipped


@pytest.mark.parametrize(
    "value, new_min, clipped",
    [
        (20, 15, Decimal(20)),
        (20, 25, Decimal(25)),
        (20, None, Decimal(20)),
        (None, None, None),
        (None, 25, None),
    ],
)
def test_clip_on_min_change(widget, value, new_min, clipped):
    """A widget's value will be clipped if the min value changes."""
    # Set an initial max, and a value that is less than it.
    widget.min = 10
    widget.value = value

    # Set a new max
    widget.min = new_min
    # Value might be clipped
    assert widget.value == clipped


def test_on_change(widget):
    """The on_change handler can be invoked."""
    # No handler initially
    assert widget._on_change._raw is None

    # Define and set a new callback
    handler = Mock()

    widget.on_change = handler

    assert widget.on_change._raw == handler

    # Invoke the callback
    widget._impl.simulate_change()

    # Callback was invoked
    handler.assert_called_once_with(widget)


@pytest.mark.parametrize(
    "value, clean",
    [
        # Valid values
        ("", ""),
        ("123", "123"),
        ("-123", "-123"),
        ("1.23", "1.23"),
        ("-1.23", "-1.23"),
        (".123", ".123"),
        # Non-alphanumeric
        ("12a3b", "123"),
        ("12!3@", "123"),
        # - not at the start
        ("1-23", "123"),
        ("123-", "123"),
        ("1.2-3", "1.23"),
        # Multiple . characters
        ("1.2.3", "1.23"),
        ("1..23", "1.23"),
        ("1.23.", "1.23"),
        ("-1.2.3", "-1.23"),
        ("-1..23", "-1.23"),
        ("-1.23.", "-1.23"),
        # Multiple problems
        ("A12.3!4.-56", "12.3456"),
    ],
)
def test_clean_decimal_str(value, clean):
    assert _clean_decimal_str(value) == clean


@pytest.mark.parametrize(
    "value, step, clean",
    [
        # Strings of integers
        ("123", None, "123"),
        ("123", "10", "123"),
        ("123", "0.01", "123.00"),
        # Strings of floats
        ("1.23456", None, "1.23456"),
        ("1.23456", "10", "1"),
        ("1.23456", "0.01", "1.23"),
        # Integers
        (123, None, "123"),
        (123, "10", "123"),
        (123, "0.01", "123.00"),
        # Floats
        (1.23456, None, "1.23456"),
        (1.23456, "10", "1"),
        (1.23456, "0.01", "1.23"),
    ],
)
def test_clean_decimal(value, step, clean):
    assert _clean_decimal(value, Decimal(step) if step else step) == Decimal(clean)


def test_deprecated_names():
    """The deprecated min_value/max_value names still work."""
    # Can't specify min and min_value
    with pytest.raises(
        ValueError,
        match=r"Cannot specify both min and min_value",
    ):
        toga.NumberInput(min=2, min_value=4)

    # Can't specify min and min_value
    with pytest.raises(
        ValueError,
        match=r"Cannot specify both max and max_value",
    ):
        toga.NumberInput(max=2, max_value=4)

    # min_value is deprecated
    with pytest.warns(
        DeprecationWarning,
        match="NumberInput.min_value has been renamed NumberInput.min",
    ):
        widget = toga.NumberInput(min_value=2)

    assert widget.min == 2

    with pytest.warns(
        DeprecationWarning,
        match="NumberInput.min_value has been renamed NumberInput.min",
    ):
        assert widget.min_value == 2

    with pytest.warns(
        DeprecationWarning,
        match="NumberInput.min_value has been renamed NumberInput.min",
    ):
        widget.min_value = 4

    assert widget.min == 4

    # max_value is deprecated
    with pytest.warns(
        DeprecationWarning,
        match="NumberInput.max_value has been renamed NumberInput.max",
    ):
        widget = toga.NumberInput(max_value=2)

    assert widget.max == 2

    with pytest.warns(
        DeprecationWarning,
        match="NumberInput.max_value has been renamed NumberInput.max",
    ):
        assert widget.max_value == 2

    with pytest.warns(
        DeprecationWarning,
        match="NumberInput.max_value has been renamed NumberInput.max",
    ):
        widget.max_value = 4

    assert widget.max == 4



class TogaApplicator:
    """Apply styles to a Toga widget."""

    def __init__(self, widget):
        self.widget = widget

    def refresh(self):
        # print("RE-EVALUATE LAYOUT", self.widget)
        self.widget.refresh()

    def set_bounds(self):
        # print("  APPLY LAYOUT", self.widget, self.widget.layout)
        self.widget._impl.set_bounds(
            self.widget.layout.absolute_content_left,
            self.widget.layout.absolute_content_top,
            self.widget.layout.content_width,
            self.widget.layout.content_height,
        )
        for child in self.widget.children:
            child.applicator.set_bounds()

    def set_text_alignment(self, alignment):
        self.widget._impl.set_alignment(alignment)

    def set_hidden(self, hidden):
        self.widget._impl.set_hidden(hidden)
        for child in self.widget.children:
            # If the parent is hidden, then so are all children. However, if the
            # parent is visible, then the child's explicit visibility style is
            # taken into account. This visibility cascades into any
            # grandchildren.
            #
            # parent hidden child hidden style child final hidden state
            # ============= ================== ========================
            # True          True               True
            # True          False              True
            # False         True               True
            # False         False              False
            child.applicator.set_hidden(hidden or child.style._hidden)

    def set_font(self, font):
        # Changing the font of a widget can make the widget change size,
        # which in turn means we need to do a re-layout
        self.widget._impl.set_font(font)
        self.widget.refresh()

    def set_color(self, color):
        self.widget._impl.set_color(color)

    def set_background_color(self, color):
        self.widget._impl.set_background_color(color)



from __future__ import annotations

from toga.handlers import wrapped_handler

from .base import Widget


class MultilineTextInput(Widget):
    def __init__(
        self,
        id=None,
        style=None,
        value: str | None = None,
        readonly: bool = False,
        placeholder: str | None = None,
        on_change: callable | None = None,
    ):
        """Create a new multi-line text input widget.

        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style
            will be applied to the widget.
        :param value: The initial content to display in the widget.
        :param readonly: Can the value of the widget be modified by the user?
        :param placeholder: The content to display as a placeholder when
            there is no user content to display.
        :param on_change: A handler that will be invoked when the value of
            the widget changes.
        """

        super().__init__(id=id, style=style)

        # Create a platform specific implementation of a MultilineTextInput
        self._impl = self.factory.MultilineTextInput(interface=self)

        # Set a dummy handler before installing the actual on_change, because we do not want
        # on_change triggered by the initial value being set
        self.on_change = None
        self.value = value

        # Set all the properties
        self.readonly = readonly
        self.placeholder = placeholder
        self.on_change = on_change

    @property
    def placeholder(self) -> str:
        """The placeholder text for the widget.

        A value of ``None`` will be interpreted and returned as an empty string.
        Any other object will be converted to a string using ``str()``.
        """
        return self._impl.get_placeholder()

    @placeholder.setter
    def placeholder(self, value):
        self._impl.set_placeholder("" if value is None else str(value))
        self.refresh()

    @property
    def readonly(self) -> bool:
        """Can the value of the widget be modified by the user?

        This only controls manual changes by the user (i.e., typing at the
        keyboard). Programmatic changes are permitted while the widget has
        ``readonly`` enabled.
        """
        return self._impl.get_readonly()

    @readonly.setter
    def readonly(self, value):
        self._impl.set_readonly(bool(value))

    @property
    def value(self) -> str:
        """The text to display in the widget.

        A value of ``None`` will be interpreted and returned as an empty string.
        Any other object will be converted to a string using ``str()``.
        """
        return self._impl.get_value()

    @value.setter
    def value(self, value):
        self._impl.set_value("" if value is None else str(value))
        self.refresh()

    def scroll_to_bottom(self):
        """Scroll the view to make the bottom of the text field visible."""
        self._impl.scroll_to_bottom()

    def scroll_to_top(self):
        """Scroll the view to make the top of the text field visible."""
        self._impl.scroll_to_top()

    @property
    def on_change(self) -> callable:
        """The handler to invoke when the value of the widget changes."""
        return self._on_change

    @on_change.setter
    def on_change(self, handler):
        self._on_change = wrapped_handler(self, handler)
